/**
  *  \file client/chartusr.cc
  *  \brief Starchart -- User Interface
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart.h"
#include "client/chart/keymapmode.h"
#include "client/chart/messagemode.h"
#include "client/chart/prefixmode.h"
#include "client/chart/standardmode.h"
#include "client/chartloc.h"
#include "client/chartmode.h"
#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dialogs/consoledlg.h"
#include "client/history.h"
#include "client/playerscr.h"
#include "client/screens.h"
#include "client/screens/ionscreen.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/lock.h"
#include "game/planet.h"
#include "game/pref.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "int/if/ifutil.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/if/userif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/keymap.h"
#include "int/process.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "client/screens/minescreen.h"
#include "client/screens/ufoscreen.h"

static GChartLocation chart_location;
static WChartModeList chart_modes;

/********************************** Main *********************************/

static void
initStarchart()
{
    chart_location.init();
}

static void
initStarchartAfterLoad()
{
    chart_location.initAfterLoad();
}

void
setStarchartPosition(GPoint pt)
{
    chart_location.setLocation(pt);
}

void
setStarchartObject(const GObject& obj)
{
    chart_location.setCurrentObject(obj);
}

GPoint
getStarchartPosition()
{
    // FIXME: simple or real canonical location?
    return chart_location.getLocation().getSimpleCanonicalLocation();
}

static const GMapObject*
doChartLock(GPoint& pt, TLockMode lm, bool marked)
{
    GLockData locker(pt, marked);
    findItem(getDisplayedTurn().getCurrentUniverse(), locker, lm);
    pt = locker.getPoint();
    return locker.getObject();
}

namespace {
    class StarchartWidget : public UIWidget, public IntUserInterfaceChildBinding {
        friend void ::doStarchart();
     public:
        StarchartWidget();
        void drawContent(GfxCanvas&);
        bool handleEvent(const UIEvent& e, bool second_pass);

        void drawChart(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo&);

        virtual IntValue* getProperty(IntUserInterfaceProperty p);
        virtual void setProperty(IntUserInterfaceProperty p, IntValue* v);
        virtual void postEvent(std::auto_ptr<IntUserInterfaceEvent> event);
        virtual void enumContexts(ptr_vector<IntContext>& ctx);

        ptr_multi_list<IntUserInterfaceEvent> eventQueue;
    };
}

StarchartWidget::StarchartWidget()
    : UIWidget(0),
      eventQueue()
{ }

void
StarchartWidget::drawContent(GfxCanvas& can)
{
    drawChart(can);
    can.update(getExtent());
}

bool
StarchartWidget::handleEvent(const UIEvent& /*e*/, bool /*second_pass*/)
{
    return true;
}

void
StarchartWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.pref_size = info.min_size = GfxPoint(100,100);
}

void
StarchartWidget::drawChart(GfxCanvas& can)
{
    GChartViewport& viewport = chart_location.getViewport();

    can.drawBar(viewport.view, standard_colors[0],
                standard_colors[0], SOLID_FILL,
                GFX_ALPHA_OPAQUE);
    viewport.init();
    chart_modes.drawBelow(can, chart_location);
    viewport.draw(can,
                  getChartOpts(false, viewport.mult, viewport.divi),
                  getDisplayedTurn().getCurrentUniverse(),
                  getDisplayedTurn().getChartBBox());
    chart_modes.drawCursor(can, chart_location);
    chart_modes.drawOverlays(can, chart_location);
}

IntValue*
StarchartWidget::getProperty(IntUserInterfaceProperty p)
{
    switch (p) {
     case iuiScreenNumber:
        return makeIntValue(4);
     case iuiScreenRegistered:
        return makeIntValue(1);
     case iuiSimFlag:
        return makeBoolValue(false);
     case iuiScanX:
     case iuiChartX:
        return makeIntValue(chart_location.getLocation().x);
     case iuiScanY:
     case iuiChartY:
        return makeIntValue(chart_location.getLocation().y);
     case iuiKeymap:
        if (IntKeymapRef view_keymap = chart_location.getKeymap()) {
            return makeKeymapValue(view_keymap);
        } else {
            return makeKeymapValue(IntKeymap::getKeymapByName("STARCHART"));
        }
    }
    return 0;
}

void
StarchartWidget::setProperty(IntUserInterfaceProperty p, IntValue* v)
{
    int32_t n;
    switch (p) {
     case iuiChartX:
     case iuiScanX:
        if (checkIntArg(n, v, 0, 10000)) {
            setStarchartPosition(GPoint(n, chart_location.getLocation().y));
        }
        break;

     case iuiChartY:
     case iuiScanY:
        if (checkIntArg(n, v, 0, 10000)) {
            setStarchartPosition(GPoint(chart_location.getLocation().x, n));
        }
        break;

     default:
        throw IntError::notAssignable();
    }
}

void
StarchartWidget::postEvent(std::auto_ptr<IntUserInterfaceEvent> event)
{
    eventQueue.push_back(event.release());
}

void
StarchartWidget::enumContexts(ptr_vector<IntContext>& ctx)
{
    IntUserInterfaceChildBinding::parent.enumContexts(ctx);
    GChartLocation& loc = chart_location;
    if (loc.getCurrentObject().isValid()) {
        if (dynamic_cast<GShip*>(&*loc.getCurrentObject())) {
            ctx.push_back(new IntShipContext(loc.getCurrentObject()->getId()));
        } else if (dynamic_cast<GPlanet*>(&*loc.getCurrentObject())) {
            ctx.push_back(new IntPlanetContext(loc.getCurrentObject()->getId()));
        } else {
            // No object
        }
    }
}

void
doStarchart()
{
    chart_location.initEntry(GfxRect(*ui_screen));
    chart_location.setCenter(chart_location.getLocation());

    StarchartWidget chart_w;
    chart_w.setExtent(chart_location.getViewport().view);
    chart_modes.deleteAllModes();
    chart_modes.pushMode(*new WStandardChartMode());

    int pending_dx = 0, pending_dy = 0;

    int speed = 0;
    int num_ticks_elapsed = 0;

    IntKeymapRef pending_keymap = 0;
    int pending_prefix = 0;

    // FIXME: handle turn change
    SignalConnection conn_univ_changed(getDisplayedTurn().getCurrentUniverse()
                                       .sig_universe_changed.add(&chart_location, &GChartLocation::markChanged));
    SignalConnection conn_univ_changed2(chart_w.observe(getDisplayedTurn().getCurrentUniverse()
                                                        .sig_universe_changed));

    ui_root->add(chart_w);
    clearSkippedTicks();
    while (1) {
        ui_root->performDeferredRedraws();

        /* Recenter map */
        if (chart_location.getCenter() != chart_location.getLocation()) {
            if (speed == 0
                && (distanceSquared(chart_location.getLocation(), chart_location.getCenter())
                    <= square(getUserPreferences().ChartAnimThreshold())))
            {
                /* Do single small movement immediately */
                chart_location.setCenter(chart_location.getLocation());
                speed = 0;
            } else {
                /* Big movement only after tick event */
                while (num_ticks_elapsed > 0) {
                    --num_ticks_elapsed;
                    GPoint a = chart_location.getLocation();
                    GPoint b = chart_location.getCenter().getSimpleNearestAlias(a);
                    double dist = std::sqrt(double(distanceSquared(a, b)));

                    if (dist <= speed) {
                        // We're very close, so go directly
                        chart_location.setCenter(chart_location.getLocation());
                        speed = 0;
                        num_ticks_elapsed = 0;
                    } else {
                        // We're a little farther. Compute speed.
                        if (dist < speed*speed) {
                            if (speed > 1)
                                --speed;
                        } else {
                            ++speed;
                        }

                        // Compute new location.
                        int distx = int((a.x - b.x) * speed / dist);
                        int disty = int((a.y - b.y) * speed / dist);

                        // Ensure we make progress.
                        if (distx == 0 && a.x != b.x)
                            distx = (a.x < b.x ? -1 : +1);
                        if (disty == 0 && a.y != b.y)
                            disty = (a.y < b.y ? -1 : +1);

                        // Move.
                        chart_location.setCenter(GPoint(b.x + distx, b.y + disty));
                    }
                }
            }
        }

        /* Process location and mode changes. Those cause a full redraw. */
        if (chart_location.isChanged() || chart_modes.isChanged()) {
            /* Let mode handle updates */
            chart_modes.onMove(chart_location);

            /* Redraw map */
            ui_screen->lock();
            chart_w.drawChart(*ui_screen);
            ui_screen->unlock();
            ui_screen->update(chart_w.getExtent());

            /* Clear change flags */
            chart_location.clearChanged();
            chart_modes.clearChanged();
            chart_modes.markClean();
        }

        /* Handle chart mode dirty regions. Those cause a partial redraw. */
        if (chart_modes.isDirty()) {
            GfxClipRectFilter filter(chart_modes.getDirtyRect(), *ui_screen);
            filter.lock();
            chart_w.drawChart(filter);
            filter.unlock();
            filter.update(chart_modes.getDirtyRect());
            chart_modes.markClean();
        }

        if (!chart_w.eventQueue.empty()) {
            std::auto_ptr<IntUserInterfaceEvent> event(chart_w.eventQueue.extract_front());
            switch (event->command) {
             case iucPopupConsole:
                doConsole();
                break;

             case iucExitRace:
             case iucExitClient:
                chart_w.IntUserInterfaceChildBinding::parent.postEvent(event);
                goto out;

             case iucOverlayMessage:
                WMessageChartMode::showMessage(chart_modes, IntValue::toStringOf(event->args[0], false));
                break;

             case iucChartSetView:
                chart_location.setViewName(IntValue::toStringOf(event->args[0], false));
                break;

             case iucGoToScreen:
                chart_w.IntUserInterfaceChildBinding::parent.postEvent(event);
                goto out;

             case iucGoToMinefield:
                doMinefieldScreen();
                break;

             case iucGoToStorm:
                doIonStormScreen();
                break;

             case iucGoToUfo:
                doUfoScreen();
                break;

             case iucGoToChart:
             case iucExitDialog:
                break;

             case iucUseKeymap:
                if (IntKeymapValue* kv = dynamic_cast<IntKeymapValue*>(event->args[0])) {
                    IntIntValue* iv = dynamic_cast<IntIntValue*>(event->args[1]);
                    pending_keymap = kv->getKeymap();
                    pending_prefix = iv != 0 ? iv->getValue() : 0;
                }
                break;
            }
            continue;
        }

        if (hasAnyRunnableProcess()) {
            runRunnableProcesses();
            killTerminatedProcesses();
            continue;
        }

        if (pending_keymap) {
            WKeymapChartMode::startKeymapMode(chart_modes, pending_keymap, pending_prefix);
            pending_keymap = 0;
        }

        /* User interface */
        UIEvent e;
        pushHistoryScreenId(WScreenId(chart_location.getLocation()));
        getEvent(e, false, true);

        /* Drive timer */
        if (e.type == ev_Tick)
            num_ticks_elapsed = e.tick.elapsed;
        else
            num_ticks_elapsed = 0;

        if (chart_modes.handleEvent(e, chart_location)) {
            /* Handled by mode */
        } else if (e.type == ev_Mouse) {
            /* Mouse movement. When mouse movement does not immediately
               translate into screen movement, we accumulate mouse movement
               in pending_d[xy] instead of discarding it. This gives much
               smoother movement in large zoom levels. */
            const GMapObject* new_obj = 0;
            pending_dx += e.mouse.x;
            pending_dy += e.mouse.y;
            GPoint new_pt = chart_location.getLocation();
            new_pt.x += chart_location.getViewport().unscale(pending_dx);
            new_pt.y -= chart_location.getViewport().unscale(pending_dy);

            if (e.mouse.b) {
                // mouse button held down: lock onto object
                if (e.mouse.b & mb_Left)
                    new_obj = doChartLock(new_pt, lm_Left, e.mouse.b & ss_Ctrl);
                else if (e.mouse.b & mb_Right)
                    new_obj = doChartLock(new_pt, lm_Right, e.mouse.b & ss_Ctrl);
            } else if (chart_location.getObjectList().size() != 0) {
                // sticky mouse
                if (square(pending_dx) + square(pending_dy) <= square(getUserPreferences().ChartMouseStickiness())) {
                    new_pt = chart_location.getLocation();
                }
            } else {
                // regular movement
            }

            if (new_pt != chart_location.getLocation()) {
                if (new_obj) {
                    chart_location.setCurrentObject(*new_obj);
                } else {
                    chart_location.setLocation(new_pt);
                }
                pending_dx = pending_dy = 0;
            }
        } else if (e.type == ev_Key) {
            /* Keys */
            if (e.key.code == ss_Ctrl + ss_Alt + 's') {
                doScreenshot();
            } else if (e.key.code >= '1' && e.key.code <= '9') {
                WPrefixChartMode::startPrefixArg(chart_modes, e.key.code - '0');
            }
        }
    }
 out:
    ui_root->close(chart_w);
    chart_modes.deleteAllModes();
}

const ModuleInitializer mod_chartusr[] = {
    { OnEnterNoGame | OnEnterDirectory | OnEnterGame, initStarchart },
    { OnEnteredGame, initStarchartAfterLoad },
    { 0, 0 },
};
