/**
  *  \file client/dialogs/outboxdlg.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/outboxdlg.h"
#include "client/widgets/msgaction.h"
#include "game/msg.h"
#include "ui/layout.h"
#include "ui/window.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "client/widgets/msgdisplay.h"
#include "ui/prefix.h"
#include "game/global.h"
#include "game/team.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/setselect.h"
#include "client/dialogs/msgreceiver.h"
#include "client/dialogbinding.h"
#include "ui/widgets/quit.h"

namespace {
    class WOutboxActionPanel : public WMessageActionPanel {
        GOutbox& outbox;
     public:
        WOutboxActionPanel(GOutbox& outbox)
            : WMessageActionPanel(outbox),
              outbox(outbox)
            {
                onBrowse();
            }
        void updateMessageControls();
        bool handleEvent(const UIEvent& event, bool second_pass);
    };

    class WOutboxWindow : public UIWindow, public IntUserInterfaceDialogBinding {
        ObjectHolder h;

        void init(GOutbox& outbox);
     public:
        WOutboxWindow(GOutbox& outbox);
    };
}

void
WOutboxActionPanel::updateMessageControls()
{
    // Outgoing controls always present
    setHaveOutgoing(true);

    // TODO: evaluate 'OBJECT' header for data transmissions when we have them
}

bool
WOutboxActionPanel::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_DELETE:
            if (UIMessageBox(_("Delete this message?"), _("Revise Messages")).doYesNoDialog()) {
                /* First, try to make sure the pointer doesn't sit on a message
                   after the new end. FIXME: can we do better? */
                int cur = outbox.getCurrent();
                int count = outbox.getCount();
                if (cur > 0 && cur+1 >= count)
                    outbox.setCurrent(count-2);

                /* Delete the message */
                outbox.deleteMessage(cur);
                if (outbox.getCount() == 0) {
                    /* If this makes the message box empty, so close the dialog.
                       Be very careful that this does not trigger a redraw of the
                       WMessageDisplay or WMessageActionPanel, which would refer
                       to the now-deleted messages. This means we cannot display
                       an "all messages have been deleted" dialog here. */
                    stop(1);
                } else {
                    /* There are some messages remaining, update display */
                    outbox.sig_sel_changed.raise();
                }
            }
            return true;

         case 't':
            {
                WPlayerSetSelector sel(0, NUM_OWNERS, 15*font_ems[FONT_NORMAL]);
                sel.setSelectedPlayers(outbox.getReceiverMask(outbox.getCurrent()));
                if (chooseMessageReceiver(sel, _("Change Receiver"), "pcc2:msgout")) {
                    /* chooseMessageReceiver ensures that the user selects at
                       least one receiver */
                    outbox.setReceiverMask(outbox.getCurrent(), sel.getSelectedPlayers());
                }
            }
            return true;

         case 'e':
         case SDLK_RETURN:
            doEditMessage(outbox, outbox.getCurrent());
            return true;
        }
    }
    return WMessageActionPanel::handleEvent(event, second_pass);
}


WOutboxWindow::WOutboxWindow(GOutbox& outbox)
    : UIWindow(_("Revise Messages"), BLUE_WINDOW, COLOR_BLACK)
{
    init(outbox);
}

void
WOutboxWindow::init(GOutbox& outbox)
{
    // HBox
    //   VBox
    //     WMessageActionPanel
    //     HBox
    //       UIButton "Close"
    //       UISpacer
    //   WMessageDisplay
    // FIXME: this dialog currently has no clickable "Help" button.
    setLayout(UIHBoxLayout::instance5);

    UIGroup& g1  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g11 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new WOutboxActionPanel(outbox)));
    g1.add(g11);

    g11.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    g11.add(h.add(new UISpacer()));

    add(g1);
    add(h.add(new WMessageDisplay(0, outbox)));
    add(h.add(new UIPrefixArg()));
    add(h.add(new WHelpWidget("pcc2:revise")));
    add(h.add(new UIQuit(1)));

    pack();
}

void
doOutboxScreen()
{
    GOutbox& outbox = getDisplayedTurn().getOutbox(getPlayerId());

    WOutboxWindow win(outbox);
    ui_root->centerWidget(win);
    win.runModal(win);
}
