/**
  *  \file client/tiles/planetoverview.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/planetoverview.h"
#include "ui/uiglobal.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/planet.h"
#include "game/tables.h"
#include "util/misc.h"
#include "game/racename.h"

/************************** WPlanetOverviewTile **************************/

WPlanetOverviewTile::WPlanetOverviewTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(font_sizes[FONT_NORMAL_BOLD].scaledBy(30, 10), selection)
{ }

void
WPlanetOverviewTile::drawData(GfxCanvas& can)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getCurrentObject());
    if (!p)
        return;

    GPlanet& pp = *p;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);

    int y = 0, x = 0;
    const int h = font_heights[FONT_NORMAL];

    /* location, climate */
    outTextF(ctx, x, y, 130, format(_("Climate: %s"), getTemperatureName(pp.getTemperature())));
    if (int sector = pp.getPos().getSectorNumber())
        outTextF(ctx, x+130, y, extent.w - 130,
                 format(_("Location: (%d,%d) - Sector %d"),
                        pp.getPos().x, pp.getPos().y, sector));
    else
        outTextF(ctx, x+130, y, extent.w - 130,
                 format(_("Location: (%d,%d)"),
                        pp.getPos().x, pp.getPos().y));
    y += h;
    y += h/3;

    // FIXME: I don't really like this because it uses hard
    // measurements (the same as in PCC 1.x)
    /*  |-35-|---70---|5|---95---|30|----175--------|
        Col:  N,NNN,NNN Race      xx% -> happy (100)
        Nat:  N,NNN,NNN Race
        |0   |35      |105|110   |205|235  */
    
    /* colonists */
    outTextF(ctx, x,   y, 35, _("Col:"));
    if (pp.getOwner()) {
        outTextF(ctx, x+110, y, 95, player_racenames.getAdjName(pp.getOwner()));
        outTextF(ctx, x+235, y, 155,
                 format(" \xE2\x86\x92 %s (%d)", getHappinessName(pp.getColonistHappiness()),
                        pp.getColonistHappiness()));
        ctx.setTextAlign(2, 0);
        outTextF(ctx, x+105, y, 70, clansToString(pp.getColonists()));
        outTextF(ctx, x+235, y, 30, format("%d%%", pp.getColonistTax()));
        ctx.setTextAlign(0, 0);
    } else {
        outTextF(ctx, x+35, y, 70+5+95+30+155, _("none"));
    }
    y += h;

    /* natives */
    outTextF(ctx, x, y, 35, _("Nat:"));
    if (pp.getNativeRace()) {
        outTextF(ctx, x+110, y, 95, getNativeRaceName(pp.getNativeRace()));
        outTextF(ctx, x+235, y, 155,
                 format(" \xE2\x86\x92 %s (%d)", getHappinessName(pp.getNativeHappiness()),
                        pp.getNativeHappiness()));
        ctx.setTextAlign(2, 0);
        outTextF(ctx, x+105, y, 70, clansToString(pp.getNatives()));
        outTextF(ctx, x+235, y, 30, format("%d%%", pp.getNativeTax()));
        ctx.setTextAlign(0 ,0);
    } else {
        outTextF(ctx, x+35, y, 70+5+95+30+155, _("none"));
    }
    y += h;
    y += h/3;

    /* industry */
    outTextF(ctx, x, y, extent.w,
             format(_("Mines: %d; Factories: %d; Defense Posts: %d"),
                    numToString(pp.getStructures(str_Mines)),
                    numToString(pp.getStructures(str_Factories)),
                    numToString(pp.getStructures(str_Defense))));
    y += h;

    outTextF(ctx, x, y, extent.w,
             format(_("%d mc; %d kt supplies"),
                    numToString(pp.getCargo(el_Money)), numToString(pp.getCargo(el_Supplies))));
    y += h;

    /* minerals */
    outTextF(ctx, x, y, 50, _("Mined:"));
    ctx.setTextAlign(2, 0);
    outTextF(ctx, x+130, y, 80, format("%dN", numToString(pp.getCargo(el_Neutronium))));
    outTextF(ctx, x+210, y, 80, format("%dT", numToString(pp.getCargo(el_Tritanium))));
    outTextF(ctx, x+290, y, 80, format("%dD", numToString(pp.getCargo(el_Duranium))));
    outTextF(ctx, x+370, y, 80, format("%dM", numToString(pp.getCargo(el_Molybdenum))));
    y += h;
    
    outTextF(ctx, x+130, y, 80, format("%dN", numToString(pp.getOreGround(el_Neutronium))));
    outTextF(ctx, x+210, y, 80, format("%dT", numToString(pp.getOreGround(el_Tritanium))));
    outTextF(ctx, x+290, y, 80, format("%dD", numToString(pp.getOreGround(el_Duranium))));
    outTextF(ctx, x+370, y, 80, format("%dM", numToString(pp.getOreGround(el_Molybdenum))));
    ctx.setTextAlign(0, 0);
    outTextF(ctx, x, y, 50, _("Ground:"));
    y += h;
    y += h/3;

    outTextF(ctx, x, y, extent.w,
             format(_("Friendly Code: %s"), pp.getFCode()));
}

