/**
  *  \file client/widgets/combatdiagram.cc
  *  \brief Combat Diagram Widget
  *
  *  This contains a WCombatDiagramWidget which displays a WCombatDiagram.
  *  This functionality is split into two classes, mainly because
  *  computing the WCombatDiagram can take time (it has to run all VCRs
  *  for classic combat) and therefore isn't very well suited for doing
  *  in a constructor.
  *
  *  This has been generalized from the original code based on classic
  *  combat to support general combat in a GVcrDatabase. It requires
  *  each fight to have at least one participant.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/combatdiagram.h"
#include "client/chart.h"
#include "client/marks.h"
#include "client/widgets/racelist.h"
#include "cpluslib/format.h"
#include "game/vcr.h"
#include "game/vcrdb.h"
#include "game/vcrobj.h"
#include "gfx/palette.h"
#include "io/io.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "util/progress.h"
#include "util/recode.h"
#include "game/racename.h"

static const int TEXT_EMS = 20;
static const int RIGHT_PX = 10;

static bool
SortGroups(const WCombatDiagram::Item& a, const WCombatDiagram::Item& b)
{
    if (a.group_id == b.group_id)
        return a.sequence < b.sequence;
    else
        return a.group_id < b.group_id;
}

/** Describe a battle. The idea is to say "<name> vs <name>" in 1:1 fights,
    and "<race> vs <race>" in fleet battles with two participating races.
    \param e the battle */
static string_t
getBattleName(GVcrEntry& e)
{
    int leftRace = 0;
    int leftSlot = 0;
    int rightRace = 0;
    int rightSlot = 0;

    for (int i = 0, n = e.getNumObjects(); i < n; ++i) {
        int pl = e.getObject(i, false).getOwner();
        if (leftRace == 0) {
            leftRace = pl;
            leftSlot = i;
        } else if (leftRace == pl) {
            leftSlot = -1;
        } else if (rightRace == 0) {
            rightRace = pl;
            rightSlot = i;
        } else if (rightRace == pl) {
            rightSlot = -1;
        } else {
            return _("Multiple races");
        }
    }

    if (leftRace == 0 || rightRace == 0) {
        return _("Unknown");
    } else {
        string_t leftName  = (leftSlot  >= 0 ? e.getObject(leftSlot,  false).getName() : player_racenames.getShortName(leftRace));
        string_t rightName = (rightSlot >= 0 ? e.getObject(rightSlot, false).getName() : player_racenames.getShortName(rightRace));
        return format("%s vs. %s", leftName, rightName);
    }
}

/***************************** WCombatDiagram ****************************/

WCombatDiagram::WCombatDiagram()
{ }

/** Initialize and build diagram. Clears a possibly previously-initialized diagram.
    \param db VCR database
    \param prog Progress monitor to watch progress */
void
WCombatDiagram::init(GVcrDatabase& db, ProgressMonitor& prog)
{
    /* Clear old data */
    lines.clear();

    /* Check all VCRs */
    int group_counter = 0;
    int sequence      = 0;
    for (int i = 0; i < db.getNumBattles(); ++i) {
        prog.tick();
        GVcrEntry& entry = db.getBattle(i);

        /* Compute result */
        entry.prepareResult(GVcrEntry::NeedQuickOutcome, prog);

        /* Assimilate first object and obtain a group Id */
        GVcrObject& left = entry.getObject(0, false);
        std::vector<WCombatDiagram::Item>::const_iterator lptr = findObject(left);
        int group_id;
        if (lptr != lines.end()) {
            /* already known, reuse group Id */
            group_id = lptr->group_id;
        } else {
            /* not known, allocate new group Id */
            group_id = ++group_counter;
            lines.push_back(Item(left.isPlanet(), left.getId(), group_id, ++sequence, i, 0));
        }

        /* Assimilate other objects */
        for (int side = 1; side < entry.getNumObjects(); ++side) {
            GVcrObject& right = entry.getObject(side, false);
            std::vector<WCombatDiagram::Item>::const_iterator rptr = findObject(right);
            if (rptr != lines.end()) {
                /* already known, rename into earlier group */
                if (rptr->group_id < group_id) {
                    renameGroup(group_id, rptr->group_id);
                    group_id = rptr->group_id;
                } else {
                    renameGroup(rptr->group_id, group_id);
                }
            } else {
                /* not known, put into this group */
                lines.push_back(Item(right.isPlanet(), right.getId(), group_id, ++sequence, i, side));
            }
        }
    }

    /* Sort */
    std::sort(lines.begin(), lines.end(), SortGroups);
}

/** Find an object in the diagram.
    \param obj Object to find
    \return iterator pointing to this object's line, end() if none */
std::vector<WCombatDiagram::Item>::const_iterator
WCombatDiagram::findObject(const GVcrObject& obj) const
{
    std::vector<Item>::const_iterator i = lines.begin();
    while (i != lines.end() && (i->planet != obj.isPlanet() || i->id != obj.getId()))
        ++i;
    return i;
}

/** Rename a group.
    \param from Replace this group Id...
    \param to ...by this one */
void
WCombatDiagram::renameGroup(int from, int to)
{
    for (std::vector<Item>::iterator i = lines.begin(); i != lines.end(); ++i)
        if (i->group_id == from)
            i->group_id = to;
}

/** Get number of lines in this diagram. */
int
WCombatDiagram::getNumLines() const
{
    return lines.size();
}

/** Get a line by number.
    \param i Line number, [0,getNumLines()) */
const WCombatDiagram::Item&
WCombatDiagram::getLine(int i) const
{
    return lines[i];
}

/** Find line number by object. The object must appear in the diagram.
    \param obj Object to find
    \return Line number, [0,getNumLines()) */
int
WCombatDiagram::findObjectLine(const GVcrObject& obj) const
{
    return findObject(obj) - lines.begin();
}

/************************** WCombatDiagramWidget *************************/

/** Constructor.
    \param db Associated VCR database
    \param diag Precomputed diagram to display */
WCombatDiagramWidget::WCombatDiagramWidget(GVcrDatabase& db, const WCombatDiagram& diag)
    : db(db),
      diag(diag),
      hover_slot(-1),
      mdown(false),
      team_colors(false),
      tt_ctl(*this),
      tt_active(false)
{ }

void
WCombatDiagramWidget::drawContent(GfxCanvas& can)
{
    /* Compute layout */
    const int left       = font_ems[FONT_SMALL]*TEXT_EMS;
    const int right      = RIGHT_PX;
    const int chartXRoom = getExtent().w - left - right;
    const int chartX     = getExtent().x + left;
    const int chartY     = getExtent().y;
    const int chartYRoom = getExtent().h;
    const int nlines     = diag.getNumLines();
    const int textHeight = font_heights[FONT_SMALL];

    /* Set up */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());

    if (hover_slot >= 0) {
        /* Highlight hover. This is the same color as used in the WScoreIconBox. */
        ctx.setRawColor(standard_colors[COLOR_GRAYSCALE+6]);

        GfxRect r(getSlotBoundingBox(hover_slot));
        if (r.h >= 2) {
            drawBar(ctx, GfxRect(r.x, r.y+1, r.w, r.h-2));
            drawHLine(ctx, r.x+1, r.y, r.x + r.w-2);
            drawHLine(ctx, r.x+1, r.y+r.h-1, r.x + r.w-2);
        }
    }

    /* Labels on left side */
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.setColor(UIColor::tc_Static);
    int nextTextY = chartY;
    for (int i = 0; i < nlines; ++i) {
        int y = chartY + chartYRoom * (2*i + 1) / (2*nlines) - textHeight/2;
        if (y >= nextTextY) {
            outTextF(ctx, getExtent().x, y, left, getObjectName(i));
            nextTextY = y + textHeight;
        }
    }

    /* Draw diagram */
    /*
     *  There is one difference between FLAK and classic combat here.
     *  Whereas classic combat reports a planet destroyed when it's captured,
     *  FLAK actually reports it captured, which means it gets its line through
     *  to the right. I'm not going to change this, though, because marking the
     *  planet destroyed in FLAK too would lose information about who destroyed
     *  (resp. captured) it, and marking it captured in classic combat would
     *  draw too many lines, making the diagram less intuitive for those.
     */
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    std::vector<int> lastRelX(nlines);
    std::vector<int> lastOwners(nlines);
    for (int i = 0, n = db.getNumBattles(); i < n; ++i) {
        /* Compute line color and figure out line range */
        GVcrEntry& entry = db.getBattle(i);
        bool anyResult = false;
        bool anyKill = false;
        int  captor = 0;
        int  minLine = 0, maxLine = 0;
        for (int side = 0, nsides = entry.getNumObjects(); side < nsides; ++side) {
            /* Locate object line */
            const int line = diag.findObjectLine(entry.getObject(side, false));
            if (side == 0 || line < minLine) {
                minLine = line;
            }
            if (side == 0 || line > maxLine) {
                maxLine = line;
            }

            /* Process status */
            int status = entry.getOutcome(side);
            if (status < 0) {
                /* killed */
                anyKill = true;
                anyResult = true;
            } else if (status > 0) {
                /* captured */
                anyResult = true;
                if (captor == 0 || captor == status) {
                    captor = status;
                } else {
                    captor = -1;
                }
            } else {
                /* survived */
            }
        }

        /* Draw vertical line */
        if (!anyResult) {
            /* nothing happened (stalemate) */
            ctx.setColor(COLOR_DARK);
            ctx.setLinePattern(DOTTED_LINE);
        } else if (!anyKill && captor > 0) {
            /* unique captor */
            ctx.setColor(getColorForOwner(captor));
        } else {
            /* non-unique captor, or destruction */
            ctx.setColor(COLOR_BLACK);
        }

        const int relX = chartXRoom * (i+1) / n;
        const int minY = chartY + chartYRoom * (2*minLine + 1) / (2*nlines);
        const int maxY = chartY + chartYRoom * (2*maxLine + 1) / (2*nlines);
        drawVLine(ctx, chartX + relX, minY, maxY);
        ctx.setLinePattern(SOLID_LINE);

        /* Draw horizontal lines and explosion markers */
        for (int side = 0, nsides = entry.getNumObjects(); side < nsides; ++side) {
            const GVcrObject& obj = entry.getObject(side, false);
            const int objLine     = diag.findObjectLine(obj);
            const int objY        = chartY + chartYRoom * (2*objLine + 1) / (2*nlines);
            const int objStatus   = entry.getOutcome(side);

            ctx.setLineThickness(3);
            ctx.setColor(getColorForOwner(obj.getOwner()));
            drawHLine(ctx, chartX + lastRelX[objLine], objY, chartX + relX);
            ctx.setLineThickness(1);

            if (objStatus < 0) {
                /* killed */
                lastOwners[objLine] = 0;
                ctx.setColor(COLOR_RED);
                drawMarker(ctx, getUserMarker(0, true), GfxPoint(chartX + relX, objY));
                ctx.setColor(COLOR_YELLOW);
                drawMarker(ctx, getUserMarker(2, true), GfxPoint(chartX + relX, objY));
            } else if (objStatus == 0) {
                /* survived */
                lastOwners[objLine] = obj.getOwner();
            } else {
                /* captured */
                lastOwners[objLine] = objStatus;
            }
            lastRelX[objLine] = relX;
        }
    }

    /* Final lines on right side */
    ctx.setLineThickness(3);
    for (int i = 0; i < nlines; ++i) {
        if (lastOwners[i] != 0) {
            const int y = chartY + chartYRoom * (2*i + 1) / (2*nlines);
            ctx.setColor(getColorForOwner(lastOwners[i]));
            drawHLine(ctx, chartX + lastRelX[i], y, getExtent().getRightX()-1);
        }
    }
    ctx.setLineThickness(1);

    can.update(getExtent());
}

void
WCombatDiagramWidget::getLayoutInfo(LayoutInfo& info)
{
    /* FIXME: we must artifically limit the pref_size because layout cannot
       yet handle pref_size exceeding screen space. */
    int addX = font_ems[FONT_SMALL]*TEXT_EMS + RIGHT_PX;
    info.min_size  = GfxPoint(std::min(400, db.getNumBattles()*5 + addX),
                              std::min(300, diag.getNumLines()*2));
    info.pref_size = GfxPoint(std::min(400, db.getNumBattles()*20 + addX),
                              std::min(300, diag.getNumLines()*10));
    info.flags     = li_Flexible;
}

bool
WCombatDiagramWidget::handleEvent(const UIEvent& event, bool second_pass)
{
    if (tt_ctl.checkEvent(event) && hover_slot >= 0) {
        GVcrEntry& e = db.getBattle(hover_slot);
        tt_active = true;
        doTooltip(format(_("Battle %d:\n%s"), hover_slot+1, getBattleName(e)),
                  tt_ctl.getTooltipPosition());
        tt_active = false;
    }

    if (event.type == ev_Mouse) {
        GfxPoint where(event.mouse.x, event.mouse.y);
        if (isInside(where)) {
            int slot = -1;
            for (int i = 0, n = db.getNumBattles(); i < n; ++i) {
                if (getSlotBoundingBox(i).contains(where)) {
                    slot = i;
                }
            }
            setSelectedWidget();
            setHoverSlot(slot);
            if (mdown && event.mouse.b == 0) {
                /* Mouse released */
                sig_click_battle.raise();
            } else {
                /* Mouse pressed */
                mdown = (event.mouse.b != 0);
            }
            return true;
        } else {
            setHoverSlot(-1);
            return UIWidget::handleEvent(event, second_pass);
        }
    } else {
        return UIWidget::handleEvent(event, second_pass);
    }
}

void
WCombatDiagramWidget::onStateChange(int state, bool enable)
{
    if ((state & st_Selected) != 0 && !enable && !tt_active) {
        mdown = false;
        setHoverSlot(-1);
    }
    tt_ctl.checkState();
    UIWidget::onStateChange(state, enable);
}

int
WCombatDiagramWidget::getHoverSlot() const
{
    return hover_slot;
}

bool
WCombatDiagramWidget::getTeamColorState() const
{
    return team_colors;
}

void
WCombatDiagramWidget::setTeamColorState(bool state)
{
    if (team_colors != state) {
        team_colors = state;
        drawWidget();
    }
}

void
WCombatDiagramWidget::setHoverSlot(int slot)
{
    if (slot != hover_slot) {
        hover_slot = slot;
        drawWidget();
    }
}

/** Get bounding box for a slot.
    \param slot Slot to compute bounding box for, [0,db.getNumBattles())
    \return Bounding box rectangle */
GfxRect
WCombatDiagramWidget::getSlotBoundingBox(int slot) const
{
    /* Compute layout */
    const int left       = font_ems[FONT_SMALL]*TEXT_EMS;
    const int right      = RIGHT_PX;
    const int chartXRoom = getExtent().w - left - right;
    const int chartX     = getExtent().x + left;
    const int chartY     = getExtent().y;
    const int chartYRoom = getExtent().h;
    const int nlines     = diag.getNumLines();
    const int n          = db.getNumBattles();

    /* Compute positions */
    GVcrEntry& entry = db.getBattle(slot);
    int topLine = diag.findObjectLine(entry.getObject(0, false));
    int botLine = topLine;
    for (int i = 1; i < entry.getNumObjects(); ++i) {
        int line = diag.findObjectLine(entry.getObject(i, false));
        if (line < topLine) {
            topLine = line;
        }
        if (line > botLine) {
            botLine = line;
        }
    }
    const int x     = chartX + chartXRoom * (slot+1) / n;
    const int topY  = chartY + chartYRoom * (2*topLine + 1) / (2*nlines);
    const int botY  = chartY + chartYRoom * (2*botLine + 1) / (2*nlines);
    const int width = std::min(chartXRoom / n, 20);
    const int yofs  = std::min(chartYRoom / (2*nlines), 10);

    /* Build result */
    GfxRect result(x - width/2, topY - yofs, width, botY - topY + 2*yofs);
    result.intersect(getExtent());
    return result;
}

string_t
WCombatDiagramWidget::getObjectName(int line) const
{
    const WCombatDiagram::Item& it = diag.getLine(line);
    const GVcrObject& obj = db.getBattle(it.appears_in).getObject(it.appears_as, false);
    if (obj.isPlanet()) {
        return format(_("%s (planet #%d)"), obj.getName(), it.id);
    } else {
        return format(_("%s (ship #%d)"), obj.getName(), it.id);
    }
}

int
WCombatDiagramWidget::getColorForOwner(int owner) const
{
    if (team_colors) {
        return GRAY_COLOR_SET[getPlayerColor(owner)];
    } else {
        return WRaceList::colors[owner - 1];
    }
}
