/**
  *  \file client/widgets/gaparamview.cc
  *  \brief Global Actions Parameter View
  *
  *  Differences to PCC 1.x version:
  *  - PCC 1.x has option labels in big font, and some
  *    help text in normal font.
  *  - PCC 1.x has an "override locks" option.
  *  - PCC 1.x uses normal/depressed buttons for options.
  *    We use the regular option format (button, option,
  *    value) also used on the simulator options.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/widgets/gaparamview.h"
#include "client/dialogs/searchdlg.h"
#include "client/globalact.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "util/nls.h"
#include "util/strtab.h"

static const char*const options[] = {
    N_("Exclude numerical FCodes"),
    N_("Exclude special FCodes"),
    N_("Objects from search result only"),
    N_("Marked objects only"),
    N_("Include ships"),
    N_("Include planets"),
    N_("Override locks"),
    0
};
static const char*const values[] = {
    N_("no"),
    N_("yes"),
    0
};
static const char f7[] = N_("F7");

/** Constructor.
    \param param Object to observe/edit */
WGlobalActionParameterView::WGlobalActionParameterView(WGlobalActionParameter& param)
    : WTile(computeSize()),
      param(param),
      btn_r(h.add(new UIButton("R", 'r'))),
      conn_changed(param.sig_changed.add(this, &WGlobalActionParameterView::onChange))
{
    init();
    onChange();
}

WGlobalActionParameterView::~WGlobalActionParameterView()
{ }

/** Compute widget size. Regular three-column layout using the big font's
    height as a basis:
    - 1x1 square for buttons
    - 0.5 space
    - option names
    - 0.5 space
    - option values (yes/no) */
GfxPoint
WGlobalActionParameterView::computeSize()
{
    const int he = font_heights[FONT_TITLE];

    int totalHeight = 8*he;
    int totalWidth =
        getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(options))
        + getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(values))
        + he*2; /* 1.5 for buttons, 0.5 for option/value */

    return GfxPoint(totalWidth, totalHeight);
}

/** Initialize. This creates/positions the child widgets. */
void
WGlobalActionParameterView::init()
{
    // Basic layout
    setLayout(LeftJustified, 0, true);

    // Add widgets
    const int he = font_heights[FONT_TITLE];
    GfxRect pos(0, 0, he-2, he-2);
    for (const char* p = "NSRM12L"; *p != 0; ++p) {
        // Add button
        UIButton& btn = (*p == 'R'
                         ? btn_r
                         : h.add(new UIButton(string_t(p, 1), std::tolower(uint8_t(*p)))));
        addControl(pos, btn);
        pos.y += he;

        // Add F7 button
        if (*p == 'R') {
            UIButton& btn_f7 = h.add(new UIButton(_(f7), SDLK_F7));
            // FIXME: magic numbers
            addControl(GfxRect(pos.x + he*3/2, pos.y, fonts[FONT_TITLE]->getTextWidth(_(f7))+5, he-2), btn_f7);
            pos.y += he;
        }
    }
}

/** Handle option change. */
void
WGlobalActionParameterView::onChange()
{
    btn_r.setState(st_Disabled, param.getSearchResult().size() == 0);
    drawWidget();
}

void
WGlobalActionParameterView::drawData(GfxCanvas& can)
{
    // We are a tile, so we draw 0-based.
    const int he = font_heights[FONT_TITLE];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    int optionX = he*3/2;
    int valueWidth = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(values));

    int optionWidth = getExtent().w - valueWidth - optionX;
    int valueX = getExtent().w;

    // Draw content
    static const uint8_t flags[] = {
        WGlobalActionParameter::ExceptNumericalFCodes,
        WGlobalActionParameter::ExceptSpecialFCodes,
        WGlobalActionParameter::OnlySearchResult,
        WGlobalActionParameter::OnlyMarked,
        WGlobalActionParameter::OnlyShips,
        WGlobalActionParameter::OnlyPlanets,
        WGlobalActionParameter::OverrideLocks
    };
    
    int y = he/2;
    for (int i = 0; options[i] != 0; ++i) {
        // Option
        ctx.setTransparentBackground();
        ctx.setTextAlign(0, 1);
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, optionX, y, optionWidth, _(options[i]));

        // Value
        ctx.setColor(UIColor::tc_Green);
        ctx.setSolidBackground();
        ctx.setTextAlign(2, 1);
        outTextF(ctx, valueX, y, valueWidth, _(values[param.hasFlag(flags[i])]));

        // Advance
        y += font_heights[FONT_TITLE];
        if (i == 2) {
            ctx.setTextAlign(0, 1);
            ctx.setColor(UIColor::tc_Static);
            outText(ctx, optionX + fonts[FONT_TITLE]->getTextWidth(_(f7)) + 10, y, _("Search..."));
            y += font_heights[FONT_TITLE];
        }
    }
}

bool
WGlobalActionParameterView::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'n':
            param.toggleFlag(param.ExceptNumericalFCodes);
            return true;
         case 's':
            param.toggleFlag(param.ExceptSpecialFCodes);
            return true;
         case 'r':
            if (param.getSearchResult().size() != 0)
                param.toggleFlag(param.OnlySearchResult);
            return true;
         case 'm':
            param.toggleFlag(param.OnlyMarked);
            return true;
         case '1':
            param.toggleFlag(param.OnlyShips);
            return true;
         case '2':
            param.toggleFlag(param.OnlyPlanets);
            return true;
         case 'l':
            param.toggleFlag(param.OverrideLocks);
            return true;
         case SDLK_F7:
            doSearch(param.getSearchResult());
            param.setFlag(param.OnlySearchResult, param.getSearchResult().size() != 0);
            onChange();
            return true;
        }
    }
    return WTile::handleEvent(event, second_pass);
}
