/**
  *  \file client/widgets/histposlist.cc
  *  \brief WHistoryShipPositionList widget
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/histposlist.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/univ.h"
#include "gfx/gfxcontext.h"
#include "util/nls.h"
#include "util/misc.h"


/** Constructor.
    \param lines Number of lines to display
    \param width Width in pixels */
WHistoryShipPositionList::WHistoryShipPositionList(int lines, int width)
    : UIListbox(0, lines, width, lines),
      ship(0),
      min_turn(0),
      max_turn(0)
{
    setShip(0);
}

WHistoryShipPositionList::~WHistoryShipPositionList()
{ }

/** Set ship to display.
    \param ship ship, can be null */
void
WHistoryShipPositionList::setShip(const GShip* ship)
{
    if (this->ship != ship) {
        this->ship = ship;
        if (ship == 0) {
            /* Make empty */
            min_turn = max_turn = 0;
        } else {
            /* Start with stored history range */
            min_turn = ship->getHistoryOldestTurn();
            max_turn = ship->getHistoryNewestTurn();

            /* Must include current turn */
            max_turn = std::max(max_turn, getCurrentTurn().getTurnNumber());

            /* Must not include negative turns */
            min_turn = std::max(min_turn, 1);

            /* Must have minimum size. This may re-enable negative turns! */
            min_turn = std::min(min_turn, max_turn - getPageSize());
        }

        setSize(max_turn - min_turn);
        setTop(max_turn - min_turn - getPageSize() + 1);

        /* Figure out position */
        int pos = max_turn - min_turn - 1;
        while (pos > 0 && !isAccessible(pos)) {
            --pos;
        }
        if (pos > 0) {
            setCurrent(pos);
        } else {
            setCurrent(max_turn - min_turn - 1);
        }

        /* Red tape */
        adjustPosition();
        sig_moved.raise(getId());
        drawWidget();
    }
}

/** Get currently-selected turn number. */
int
WHistoryShipPositionList::getTurn() const
{
    return min_turn + getCurrent() + 1;
}

/** Select a turn number. */
void
WHistoryShipPositionList::setTurn(int turn)
{
    scrollTo(turn - 1 - min_turn, 1);
}

void
WHistoryShipPositionList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    // Allocate space: 3 em for turn, 5 em for mass, remainder for name
    int space = getExtent().w;
    int turnWidth = std::min(space, 3*font_ems[FONT_NORMAL]);
    space -= turnWidth;
    int massWidth = std::min(space, 5*font_ems[FONT_NORMAL]);
    space -= turnWidth;
    int posWidth = space;

    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (ship) {
            int trn = min_turn + i + getTop() + 1;
            if (trn > 0) {
                const TDbShipTrackEntry& track = ship->getHistoryEntry(trn);

                // Turn number
                ctx.setTextAlign(2, 0);
                outTextF(ctx, area.x+turnWidth, area.y, turnWidth, format("%d: ", trn));

                // Position
                ctx.setTextAlign(0, 0);
                outTextF(ctx, area.x+turnWidth, area.y, posWidth,
                         track.x >= 0 && track.y >= 0
                         ? getDisplayedTurn().getCurrentUniverse().getLocationName(GPoint(track.x, track.y), locs_Orbit | locs_WW)
                         : _("unknown"));

                // Mass
                if (track.mass > 0) {
                    ctx.setTextAlign(2, 0);
                    outTextF(ctx, area.x + area.w, area.y, massWidth, format(_("%d kt"), numToString(track.mass)) + " ");
                }
            }
        }
    }
}

bool
WHistoryShipPositionList::isAccessible(int pos)
{
    if (!ship) {
        return false;
    }

    int trn = min_turn + pos + 1;
    if (trn <= 0) {
        return false;
    }

    const TDbShipTrackEntry& track = ship->getHistoryEntry(trn);
    if (track.x < 0 || track.y < 0) {
        return false;
    }
    return true;
}
