/**
  *  \file game/actions/planetstructs.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/planetstructs.h"
#include "util/nls.h"
#include "game/planetform.h"
#include "game/actions/preconditions.h"

/********************** GPlanetBuildStructuresAction *********************/

/*! \class GPlanetBuildStructuresAction

    This class allows you to build structures on a planet. The cost
    can be billed to any container. The class enforces that maximum
    limits are never exceeded (i.e. it will never go into an error
    state because of exceeded building limits, it will simply never
    let you build so much), but you can build more than you have money
    for (to compute the price).

    \todo currently, we're storing /max/ internally. If someone
    modifies the planet inbetween (i.e. add or remove colonists), we
    will not immediately adapt to that. However, this will not violate
    rules. Maybe we should change /max/ to yield the value used by
    host, i.e. sum of all colonists at this place, not just on the
    planet. */

/** Constructor.
    \param the_planet planet we are building on
    \param container  will pay for the structures
    \pre we are playing the planet and the container */
GPlanetBuildStructuresAction::GPlanetBuildStructuresAction(GPlanet& the_planet,
                                                           GCargoContainer& container)
    : GCargoCostTransaction(container),
      the_planet(the_planet),
      conn(the_planet.sig_changed.add(this, &GPlanetBuildStructuresAction::update))
{
    mustBePlayed(the_planet);
    for (int i = 0; i < MAX; ++i) {
        TPlanetaryStructure ps = TPlanetaryStructure(i);
        data[ps].order = data[ps].min = the_planet.getStructures(ps);
        data[ps].max = getMaxStructures(the_planet, ps);
        /* we may already have more than permitted by the rules, so assume it
           ended up there in a legal way (e.g. by beaming off clans). */
        if (data[ps].max < data[ps].order)
            data[ps].max = data[ps].order;
    }
    update();  /* should not be necessary, will set the cost to 0 */
}

/** Set undo information. Enables this transaction to undo former
    GPlanetBuildStructuresAction's. */
void
GPlanetBuildStructuresAction::setUndoInformation(const GPlanet& dis_planet)
{
    for (int i = 0; i < MAX; ++i) {
        mp16_t dis_count = dis_planet.getStructures(TPlanetaryStructure(i));
        if (dis_count.isKnown() && dis_count < data[i].min) {
            data[i].min = dis_count;
        }
    }
}

/** Destructor. */
GPlanetBuildStructuresAction::~GPlanetBuildStructuresAction()
{ }

/** Recompute costs.
    \todo This could be a little smarter about changes from the
    outside: never turn a build order into a scrap order (build 3
    through this instance, build 5 externally -> this instance will
    now have a "scrap 2" order), update max values. */
void
GPlanetBuildStructuresAction::update()
{
    if (!isFinished()) {
        const int nmines = data[str_Mines].order       - the_planet.getStructures(str_Mines);
        const int nfact  = data[str_Factories].order   - the_planet.getStructures(str_Factories);
        const int ndef   = data[str_Defense].order     - the_planet.getStructures(str_Defense);
        const int nsbd   = data[str_BaseDefense].order - the_planet.getStructures(str_BaseDefense);

        GCost cost;
        cost.set(el_Money,    4 * nmines + 3 * nfact + 10 * (ndef + nsbd));
        cost.set(el_Supplies, nmines + nfact + ndef);
        cost.set(el_Duranium, nsbd);
        setCost(cost);

        sig_changed.raise();
    }
}

/** Add structures. This routine will not add or remove more than allowed by the
    rules, but it can add more structures than there is cash for. In that case,
    the GPlanetBuildStructuresAction will go into error state.
    \param ps      structure type
    \param count   amount to add (negative: remove)
    \param partial allow partial add

    \return number of structures added. With partial=false, return
    value is guaranteed to be either 0 or count. */
int
GPlanetBuildStructuresAction::addStructures(TPlanetaryStructure ps, int count, bool partial)
{
    if (count > 0) {
        if (data[ps].order + count > data[ps].max) {
            /* request may exceed limit */
            if (!partial)
                return 0;
            count = data[ps].max - data[ps].order;
        }
    } else {
        /* remove things */
        if (data[ps].order + count < data[ps].min) {
            if (!partial)
                return 0;
            count = data[ps].min - data[ps].order;
        }
    }

    if (count != 0) {
        data[ps].order += count;
        update();
    }

    return count;
}

/** Add structures limited by cash. Like addStructures(), but will not add more than
    resources allow. If already over-allocated, this will not automatically turn an
    'add' order into a removal.
    \param ps      structure type
    \param count   amount to add (negative to remove)
    \return number of structures built */
int
GPlanetBuildStructuresAction::addStructuresLimitCash(TPlanetaryStructure ps, int count)
{
    /* check how much we can add according to the rules */
    int rule_count = addStructures(ps, count, true);

    /* okay, we can add that much, but we don't have the cash. */
    while (rule_count > 0 && !isValidCost()) {
        addStructures(ps, -1, true);
        rule_count--;
    }
    return rule_count;
}

/** Autobuild. Performs standard auto-build operation, and leaves result in
    this transaction. You must still commit it. */
void
GPlanetBuildStructuresAction::doStandardAutoBuild()
{
    /* If planet does not have a factory, but wants some, start by building one
       first, independant of orders */
    if (data[str_Factories].order == 0 && the_planet.getAutobuildGoal(str_Factories) != 0) {
        if (!addStructuresLimitCash(str_Factories, 1)) {
            /* we're unable to build anything */
            return;
        }
    }

    /* Figure out order in which to build */
    TPlanetaryStructure order[MAX];
    for (int i = 0; i < MAX; ++i)
        order[i] = TPlanetaryStructure(i);
    for (int i = 0; i < MAX; ++i)
        for (int j = 0; j < MAX-1; ++j)
            if (the_planet.getAutobuildSpeed(order[j]) < the_planet.getAutobuildSpeed(order[j+1]))
                std::swap(order[j], order[j+1]);

    /* Do it */
    bool have_built;
    do {
        have_built = false;
        int index = 0;
        while (index < MAX) {
            /* Find out how many items have the same speed */
            int speed = the_planet.getAutobuildSpeed(order[index]);
            int group = 1;
            while (index + group < MAX && speed == the_planet.getAutobuildSpeed(order[index+group]))
                ++group;

            /* Build these */
            for (int n = 0; n < speed; ++n) {
                bool did = false;
                for (int i = 0; i < group; ++i) {
                    const TPlanetaryStructure what = order[index+i];
                    if (data[what].order < the_planet.getAutobuildGoal(what))
                        if (addStructuresLimitCash(order[index+i], 1))
                            did = true, have_built = true;
                }
                if (!did)
                    break;
            }

            index += group;
        }
    } while (have_built);
}

/** Commit transaction. */
void
GPlanetBuildStructuresAction::commit()
{
    ASSERT(!getError());
    for (int i = 0; i < MAX; ++i) {
        ASSERT(data[i].order <= data[i].max);
        ASSERT(data[i].order >= data[i].min);
        the_planet.setStructures(TPlanetaryStructure(i), mp16_t(data[i].order));
    }
    GCargoCostTransaction::commit();
}
