/**
  *  \file game/alliance.cc
  *  \brief Alliance Handling Core
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stddef.h>
#include "game/alliance.h"
#include "cpluslib/assert.h"
#include "game/struct.h"


static void
mergeSet(GPlayerArray<GAllianceOffer::OfferType>& dst, const GPlayerArray<GAllianceOffer::OfferType>& src)
{
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (src[i] != GAllianceOffer::Unknown) {
            dst[i] = src[i];
        }
    }
}

/***************************** GAllianceLevel ****************************/

/** Define an alliance level.
    \param name Name for user (translated)
    \param id Identifier for program use
    \param flags Flags */
GAllianceLevel::GAllianceLevel(const char* name, const char* id, int flags)
    : name(name),
      id(id),
      flags(flags)
{ }

GAllianceLevel::~GAllianceLevel()
{ }

/***************************** GAllianceOffer ****************************/

/** Constructor.
    Make blank (unknown) alliance offer. */
GAllianceOffer::GAllianceOffer()
{
    theirOffer.setAll(Unknown);
    oldOffer.setAll(Unknown);
    newOffer.setAll(Unknown);
}

/** Merge two alliance offer sets.
    Values from the other set are copied into this one, excluding Unknown values.. */
void
GAllianceOffer::merge(const GAllianceOffer& other)
{
    mergeSet(theirOffer, other.theirOffer);
    mergeSet(oldOffer, other.oldOffer);
    mergeSet(newOffer, other.newOffer);
}

/** Check whether the given OfferType means a positive offer. */
bool
GAllianceOffer::isOffer(OfferType type)
{
    return type == Yes || type == Conditional;
}

/******************************* GAlliances ******************************/

/** Constructor. */
GAlliances::GAlliances()
    : levels(),
      offers(),
      handlers()
{ }

/** Copy constructor.
    Makes a new GAlliances object containing the same alliance levels and offers, but no change handlers.
    You can modify the copy and write it back using copyFrom(). */
GAlliances::GAlliances(const GAlliances& other)
    : levels(other.levels),
      offers(other.offers),
      handlers()
{ }

/** Destructor. */
GAlliances::~GAlliances()
{ }

/** Postprocess after game load.
    Calls all registered handlers. */
void
GAlliances::postprocess()
{
    for (size_t i = 0; i < handlers.size(); ++i) {
        handlers[i]->postprocess(*this);
    }
}

/** Add a new alliance level.
    \param level Level description */
void
GAlliances::addLevel(const GAllianceLevel& level)
{
    levels.push_back(level);
    offers.push_back(GAllianceOffer());
}

/** Merge from another alliance object.
    Modifies all offers to the same as in the other object.
    This is an intelligent merge that can deal with different structures on both sides. */
void
GAlliances::copyFrom(const GAlliances& other)
{
    for (size_t i = 0; i < other.levels.size(); ++i) {
        index_t index = find(other.levels[i].getId());
        if (index != nil) {
            offers[index] = other.offers[i];
        }
    }
    callHandlers();
}

/** Find an alliance level, by Id.
    \param id Id to look for
    \return index, nil if not found */
GAlliances::index_t
GAlliances::find(const string_t& id) const
{
    for (size_t i = 0, e = levels.size(); i < e; ++i) {
        if (levels[i].getId() == id) {
            return i;
        }
    }
    return nil;
}

/** Get level by index.
    \param index index, possibly obtained from find()
    \return level description, null if index is out of range */
const GAllianceLevel*
GAlliances::getLevel(index_t index) const
{
    return index < levels.size()
        ? &levels[index]
        : 0;
}

/** Get offer by index.
    \param index index, possibly obtained from find()
    \return offer, null if index is out of range */
const GAllianceOffer*
GAlliances::getOffer(index_t index) const
{
    return index < offers.size()
        ? &offers[index]
        : 0;
}

/** Get mutable offer by index.
    This method is for use by implementations of process() only.
    Normal manipulation should use the set(), setAll(), and copyFrom() methods.
    \param index index, possibly obtained from find()
    \return offer, null if index is out of range */
GAllianceOffer*
GAlliances::getMutableOffer(index_t index)
{
    return index < offers.size()
        ? &offers[index]
        : 0;
}

/** Check for offer by type.
    Checks whether there is any positive offer to or from the specified player of a level defined by the given flag.
    This can be used to give a quick overview: "there is an alliance".
    \param player Player to check
    \param flag Flag to check (IsOffer, IsEnemy)
    \param fromUs true to check for offers from us, false to check for offers to us
    \retval true there is at least one such offer
    \retval false there is no such offer */
bool
GAlliances::isAny(int player, int flag, bool fromUs) const
{
    for (size_t i = 0, e = offers.size(); i < e; ++i) {
        if (levels[i].hasFlag(flag)
            && GAllianceOffer::isOffer(fromUs ? offers[i].newOffer[player] : offers[i].theirOffer[player]))
        {
            return true;
        }
    }
    return false;
}

/** Set all offers by type.
    Sets all offers to the specified player for all levels defined by the given flag.
    This can be used to quickly set a set of levels without specifying its identifier.
    
    \param player Player to modify
    \param flag Flag to check (IsOffer, NeedsOffer, IsEnemy)
    \param set true to set negative offers (Unknown, No) to positive (Yes).
               false to set positive offers (Yes, Conditional) to negative (No). */
void
GAlliances::setAll(int player, int flag, bool set)
{
    bool did = false;
    for (size_t i = 0, e = offers.size(); i < e; ++i) {
        if (levels[i].hasFlag(flag) && GAllianceOffer::isOffer(offers[i].newOffer[player]) != set) {
            offers[i].newOffer[player] = set ? GAllianceOffer::Yes : GAllianceOffer::No;
            did = true;
        }
    }
    if (did) {
        callHandlers();
    }
}

/** Set a single alliance offer.
    Unlike direct access to getMutableOffer(), this will call listeners.
    \param index Index obtained by find()
    \param player Player to offer
    \param type New setting */
void
GAlliances::set(index_t index, int player, GAllianceOffer::OfferType type)
{
    if (index < offers.size()
        && GPlayerArray<GAllianceOffer::OfferType>::isValidIndex(player)
        && type != offers[index].newOffer[player])
    {
        // We are changing a valid offer.
        offers[index].newOffer[player] = type;

        // Notify listeners to update downstream data.
        callHandlers();
    }
}

void
GAlliances::callHandlers()
{
    for (size_t i = 0; i < handlers.size(); ++i) {
        handlers[i]->handleChanges(*this);
    }
}
