/**
  *  \file game/fleetfc.cc
  *  \brief Class GFleetFCodeChanger
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstddef>
#include "game/fleetfc.h"
#include "game/fcode.h"
#include "game/ship.h"
#include "game/univ.h"
#include "game/global.h"

/** Constructor.
    \param trn Turn to work on
    \param fleet_id Fleet Id */
GFleetFCodeChanger::GFleetFCodeChanger(GGameTurn& trn, int fleet_id)
    : trn(trn),
      fleet_id(fleet_id),
      info()
{
    init();
}

GFleetFCodeChanger::~GFleetFCodeChanger()
{ }

/** Set friendly code on all fleet members to the same value.
    \param fcode New fcode */
void
GFleetFCodeChanger::setFCode(const string_t fcode)
{
    GUniverse& univ = trn.getCurrentUniverse();
    for (std::size_t i = 0; i < info.size(); ++i) {
        univ.getShip(info[i].sid).setFCode(fcode);
    }
}

/** Set friendly code on all fleet members to avoid a same value.
    Attempts to set the fcodes such that no fleet member has the specified code.
    \param avoid Code to avoid */
void
GFleetFCodeChanger::unsetFCode(const string_t avoid)
{
    GUniverse& univ = trn.getCurrentUniverse();
    GUniverse& prev = trn.getPreviousUniverse();
    for (std::size_t i = 0; i < info.size(); ++i) {
        int sid = info[i].sid;
        string_t fc = info[i].old_fcode;
        if (fc == avoid) {
            fc = prev.getShip(sid).getFCode();
        }
        if (fc == avoid) {
            fc = info[i].random_fcode;
        }
        univ.getShip(sid).setFCode(fc);
    }
}

/** Undo all friendly code changes.
    Restores the fcodes that were set before this object was constructed. */
void
GFleetFCodeChanger::undo()
{
    GUniverse& univ = trn.getCurrentUniverse();
    for (std::size_t i = 0; i < info.size(); ++i) {
        univ.getShip(info[i].sid).setFCode(info[i].old_fcode);
    }
}

/** Initialize. Set up data structures. */
void
GFleetFCodeChanger::init()
{
    const GUniverse& univ = trn.getCurrentUniverse();
    for (int sid = univ.ty_played_ships.findNextIndex(0); sid != 0; sid = univ.ty_played_ships.findNextIndex(sid)) {
        if (sid == fleet_id || univ.getShip(sid).getFleetNumber() == fleet_id) {
            info.push_back(Info(sid, univ.getShip(sid).getFCode(), GFCode::generateRandomFCode()));
        }
    }
}
