/**
  *  \file game/hullfunc.h
  *  \brief Hull Function Management
  *
  *  This code is a little more flexible than it needs to be, to be
  *  open to support a future DEVICES.INI implementation. Not only the
  *  hull function assignments, also their definitions are taken from
  *  a text file. This also allows us to attach some useful
  *  information; see GBasicHullFunction.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_HULLFUNC_H
#define PCC_V2_GAME_HULLFUNC_H

#include <vector>
#include <utility>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "game/exp.h"
#include "game/playerset.h"
#include "cpluslib/ptrvector.h"

class GHull;
class GGameConfig;

/** Standard hull functions numbers. These are used by our core to
    check for presence of specific ship abilities. We do not require
    actual hull function numbers to be one of the hf_XXX constants. */
enum {
    hf_MerlinAlchemy,             ///< 0 = Merlin Alchemy: 9 Sup -> 3 Min
    hf_NeutronicRefinery,         ///< 1 = Neutronic Refinery: Min + Sup -> Fuel
    hf_AriesRefinery,             ///< 2 = Aries Refinery: Min -> Fuel
    hf_HeatsTo50,                 ///< 3 = Bohemian Terraformer
    hf_CoolsTo50,                 ///< 4 = Eros Terraformer
    hf_HeatsTo100,                ///< 5 = Onyx Terraformer
    hf_Hyperdrive,                ///< 6
    hf_Gravitonic,                ///< 7
    hf_ScansAllWormholes,         ///< 8 = Bohemian
    hf_LadyRoyale,                ///< 9 = Lady Royale
    hf_LokiAnticloak,             ///< 10 = Loki
    hf_ImperialAssault,           ///< 11 = SSD
    hf_FirecloudChunnel,          ///< 12 = Firecloud
    hf_Ramscoop,                  ///< 13 = Cobol
    hf_FullBioscan,               ///< 14 = Pawn
    hf_AdvancedCloak,             ///< 15 = Dark Wing
    hf_Cloak,                     ///< 16
    hf_Bioscan,                   ///< 17
    hf_SaberGlory,                ///< 18 = Saber (10% damage to own ships)
    hf_D19bGlory,                 ///< 19 = D19b (20% damage to own ships)
    hf_Unclonable,                ///< 20
    hf_CloneOnce,                 ///< 21
    hf_Ungiveable,                ///< 22
    hf_GiveOnce,                  ///< 23
    hf_Level2Tow,                 ///< 24
    hf_Tow,                       ///< 25 = depends on AllowOneEngineTowing setting
    hf_ChunnelSelf,               ///< 26
    hf_ChunnelOthers,             ///< 27
    hf_ChunnelTarget,             ///< 28
    hf_PlanetImmunity,            ///< 29 = Rebels, Klingons, if configured, plus SSD
    hf_OreCondenser,              ///< 30
    hf_Boarding,                  ///< 31 = Privs, Crystals
    hf_AntiCloakImmunity,         ///< 32 = implied by AntiCloakImmunity option
    hf_Academy,                   ///< 33
    hf_Repairs,                   ///< 34
    hf_FullWeaponry,              ///< 35 = Feds, if configured
    hf_HardenedEngines,           ///< 36
    hf_Commander,                 ///< 37
    hf_IonShield,                 ///< 38
    hf_HardenedCloak,             ///< 39
    hf_AdvancedAntiCloak          ///< 40
};

/** Basic hull function. This defines a basic hull function as known to
    the host. We read their definition from a definition file, to allow
    easier upgrade (and storage of friendly help texts) for future
    functions. */
class GBasicHullFunction {
    int id;
    string_t name;
    string_t description;
    string_t explanation;
    string_t picture_name;
    int implied_device;
 public:
    GBasicHullFunction(int id, string_t name);
    ~GBasicHullFunction();

    void setName(const string_t name);
    void setDescription(const string_t desc);
    void setExplanation(const string_t explanation);
    void addToExplanation(const string_t explanation);
    void setImpliedDevice(int implied_device);
    void setPictureName(const string_t& name);

    int getId() const;
    const string_t& getName() const;
    const string_t& getDescription() const;
    const string_t& getExplanation() const;
    int getImpliedDevice() const;
    const string_t& getPictureName() const;

    int getDamageLimit(int forOwner, const GGameConfig& config) const;
};

/** Hull function. This class has two purposes:
    - it can store a restricted/modified function definition for use
      in mapping our internal numbers / host's numbers to functions
    - it can report a broken-down function assignment for users

    In the first case, the fields Levels, HostId, BasicFunctionId are
    relevant, in the second case the fields Levels, Players, Kind, and
    BasicFunctionId.

    \see GHullFunctionData. */
class GHullFunction {
 public:
    enum Kind {
        AssignedToShip,
        AssignedToHull,
        AssignedToRace
    };

    GHullFunction(int basic_function_id = -1);
    GHullFunction(int basic_function_id, GExpLevelSet levels);

    void setPlayers(GPlayerSet players);
    void setLevels(GExpLevelSet levels);
    void setKind(Kind kind);
    void setHostId(int host_id);
    void setBasicFunctionId(int basic_function_id);

    GPlayerSet getPlayers() const;
    GExpLevelSet getLevels() const;
    Kind getKind() const;
    int getHostId() const;
    int getBasicFunctionId() const;

    bool isSame(const GHullFunction& other);

 private:
    int            basic_function_id;
    GPlayerSet     players;
    GExpLevelSet   levels;
    Kind           kind;
    int            host_id;
};

/** Hull function specification data. An object of this type contains all
    specification data related to hull functions. That is
    - description of the functions, including standard assignments
    - definition of modified functions
    - current assignments (hullfunc.txt)

    We map all functions to integers (dev_t). Although we currently
    maintain a 1:1 relationship between basic function numbers and our
    numbers for unrestricted functions, user code should avoid using
    that property.

    To initialize this object for a blank game directory, call
    loadBasicFunctionDefinitions(). To initialize it for a game, also
    use loadBasicFunctionDefinitions(), and then use getIdFromFunction()
    using the data found in the game's util.dat or hullfunc.txt to fill
    in the rest. */
class GHullFunctionData {
    GHullFunctionData(GHullFunctionData&);
    GHullFunctionData& operator=(GHullFunctionData&);

    /** All basic hull functions. This is the definition of the hull functions
        as read from specs/hullfunc.cc. It contains all names, identifiers and
        descriptions. */
    ptr_vector<GBasicHullFunction> basic_functions;

    /** Modified hull functions. This defines the modified (=level-restricted)
        hull functions. The \c players and \c kind fields are irrelevant here
        and remain at defaults. */
    ptr_vector<GHullFunction> modified_functions;

    /** Default assignments. Contains a list of pairs (hull,basicFunctionId). */
    std::vector<std::pair<int,int> > default_assignments;

    enum { MODIFIED_FUNCTION_BASE_ID = 99999 };

 public:
    // Use this type definition to check for type mismatches:
    // enum dev_t { dev_t_min = -1, dev_t_max = 999999 };
    typedef int32_t dev_t;

    GHullFunctionData();
    ~GHullFunctionData();

    // Initialisation
    void clear();
    void loadBasicFunctionDefinitions();

    // dev_t operations
    dev_t getIdFromHostId(int host_id) const;
    dev_t getIdFromFunction(const GHullFunction& def);

    // GHullFunction operations
    bool getFunctionDefinition(dev_t id, GHullFunction& asgn) const;

    // GBasicHullFunction operations
    const GBasicHullFunction* getBasicFunctionById(int id) const;
    const GBasicHullFunction* getBasicFunctionByName(string_t name, bool partial = false) const;
    GBasicHullFunction* addBasicFunction(int id, string_t name);
    bool matchBasicFunction(int requested_id, int found_id) const;

    // Default assignments
    void addDefaultAssignment(int hull_id, int basic_function_id);
    void performDefaultAssignments(ptr_vector<GHull>&) const;
    GPlayerSet getDefaultAssignment(dev_t device, const GHull& hull) const;


    static GHullFunctionData instance;
};

/** List of hull functions. This wraps a vector-of-GHullFunction, and offers
    nice useful operations on it. */
class GHullFunctionList {
    typedef std::vector<GHullFunction> container_t;
    container_t data;

 public:
    typedef container_t::const_iterator iterator;

    GHullFunctionList();
    ~GHullFunctionList();

    void add(const GHullFunction& f);

    // Container accessor interface
    container_t::size_type size() const;
    iterator begin() const;
    iterator end() const;
    void clear();
    const GHullFunction& operator[](container_t::size_type) const;

    // Operations
    void simplify();
    void sortForNewShip(GPlayerSet forPlayer);
};


/******************************** Inlines  *******************************/

/** Set player restriction.
    \param players Player set */
inline void
GHullFunction::setPlayers(GPlayerSet players)
{
    this->players = players;
}

/** Set level restriction.
    \param levels Level set */
inline void
GHullFunction::setLevels(GExpLevelSet levels)
{
    this->levels = levels;
}

/** Set kind. This describes how this function is assigned to a ship.
    \param kind Kind of assignment */
inline void
GHullFunction::setKind(Kind kind)
{
    this->kind = kind;
}

/** Set host Id. This is the number under which a level-restricted function
    is known to the host. */
inline void
GHullFunction::setHostId(int host_id)
{
    this->host_id = host_id;
}

/** Set basic function Id.
    \param basic_function_id Function Id, hf_XXX */
inline void
GHullFunction::setBasicFunctionId(int basic_function_id)
{
    this->basic_function_id = basic_function_id;
}

/** Get player restriction. */
inline GPlayerSet
GHullFunction::getPlayers() const
{
    return players;
}

/** Get level restriction. */
inline GExpLevelSet
GHullFunction::getLevels() const
{
    return levels;
}

/** Get kind of assignment. */
inline GHullFunction::Kind
GHullFunction::getKind() const
{
    return kind;
}

/** Get host Id.
    \return host id, or -1 if none */
inline int
GHullFunction::getHostId() const
{
    return host_id;
}

/** Get basic function Id. */
inline int
GHullFunction::getBasicFunctionId() const
{
    return basic_function_id;
}

#endif
