/**
  *  \file game/univ.cc
  *  \brief Universe: A Container for (Almost) Everything
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/univ.h"
#include "game/planet.h"
#include "cpluslib/misc.h"
#include "game/ship.h"
#include "game/ion.h"
#include "game/host.h"
#include "game/formula.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/mine.h"
#include "util/misc.h"
#include "game/parser.h"
#include "game/stat.h"
#include "game/mission.h"
#include "game/config.h"


/** Update all objects in a particular object type.
    \param type The type to update */
static bool
updateType(GObjectType& type)
{
    bool changed = false;
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        GObject& obj = type.getObjectByIndex(i);
        if (obj.isDirty()) {
            obj.notifyListeners();
            changed = true;
        }
    }
    return changed;
}

/** Set playability of an object type. */
static void
setTypePlayability(GObjectType& type, GObject::Playability playability)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        type.getObjectByIndex(i).setPlayability(playability);
    }
}


/** Create blank universe. */
GUniverse::GUniverse()
    : ty_played_ships(*this),
      ty_any_ships(*this),
      ty_history_ships(*this),
      ty_played_planets(*this),
      ty_any_planets(*this),
      ty_played_bases(*this),
      ty_fleets(*this),
      ty_ufos(*this),
      ty_ion_storms(*this),
      ty_minefields(*this),
      universe_changed(false),
      turn_nr(0)
{
    init();
}

/** Destroy universe. */
GUniverse::~GUniverse()
{ }

/** Postprocess universe. Call this to make sure structural changes propagate.
    \param playingSet Set of players we play
    \param dataSet Set of players we have data for
    \param playability Playability to set for things we "play" */
void
GUniverse::postprocess(const GPlayerSet playing_set, GPlayerSet data_set, GObject::Playability playability)
{
    this->data_set = data_set;

    // Internal check for ships and planets
    // FIXME: normally, we should make units that we have full data for but are not playing ReadOnly.
    // It happens that data arriving this way may actually be partial (c2nu allied planets)
    // and much code cannot deal with this partial data.
    // Because we don't have any way of setting up a game with playing_set differing from data_set anyway,
    // blocking this completely does not make much of a difference.
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = getPlanet(i);
        p.internalCheck();
        p.setPlayability(p.isVisible() && p.hasFullPlanetData() && p.getOwner() != 0
                         ? (playing_set.contains(p.getOwner())
                            ? playability
                            : GObject::NotPlayable /*ReadOnly*/)
                         : GObject::NotPlayable);
    }
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getShip(i);
        s.internalCheck();
        s.setPlayability(s.isVisible() && s.hasFullShipData()
                         ? (playing_set.contains(s.getOwner())
                            ? playability
                            : GObject::NotPlayable /*ReadOnly*/)
                         : GObject::NotPlayable);
    }

    // Internal checks for others
    ty_minefields.internalCheck(getTurnNumber());
    drawings.eraseExpiredDrawings(getTurnNumber());
    ty_ufos.postprocess();
    setTypePlayability(ty_minefields, playability);
    setTypePlayability(ty_ufos, playability);

    // Combined checks
    for (int i = 1; i <= NUM_SHIPS; ++i)
        getShip(i).combinedCheck1(*this);
    for (int i = 1; i <= NUM_PLANETS; ++i)
        getPlanet(i).combinedCheck2(*this);

    // FIXME: synthesize scores if a score blanker is used
}

/** Recompute planet orbit flags.
    This regenerates the color flags after a possible change of team relations,
    but does not recompute the positions. */
void
GUniverse::recomputeOrbitFlags()
{
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        getPlanet(i).clearOrbitFlags();
    }

    for (int i = ty_any_ships.findNextIndex(0); i != 0; i = ty_any_ships.findNextIndex(i)) {
        const GShip& sh = getShip(i);
        if (sh.getOrbitPlanetId() != 0) {
            getPlanet(sh.getOrbitPlanetId()).setShipInOrbit(sh);
        }
    }
}

/** Get planet at location, with warp wells. If there is no planet at
    the location, look whether the point is in the warp well of one.
    \param pt Location, need not be normalized
    \param gravity true to look into warp well. If false, function
                   behaves exactly like getPlanetAt(int).
    \return Id of planet, or zero if none */
int
GUniverse::getPlanetAt(const GPoint pt, bool gravity) const
{
    int rv = getPlanetAt(pt);
    if (!rv && gravity)
        rv = getGravityPlanetAt(pt);
    return rv;
}

/** Get planet at location.
    \param pt Location, need not be normalized
    \return Id of planet, or zero if none */
int
GUniverse::getPlanetAt(const GPoint pt) const
{
    const GPoint real_pos = pt.getCanonicalLocation();
    for (int i = ty_any_planets.findNextIndex(0); i != 0; i = ty_any_planets.findNextIndex(i)) {
        if (getPlanet(i).getPos() == real_pos) {
            return i;
        }
    }
    return 0;
}

/** Get planet from warp well location.
    \param pt Location
    \pre getPlanetAt(pt) == 0
    \return Id of planet if pt is in its warp wells, 0 otherwise */
int
GUniverse::getGravityPlanetAt(GPoint pt) const
{
    /* easy case */
    if (!config.AllowGravityWells())
        return 0;

    if (host.isPHost()) {
        /* PHost gravity wells */
        int sqs = square(config.GravityWellRange());
        for (int i = NUM_PLANETS; i > 0; --i) {
            if (ty_any_planets.isValidIndex(i)) {
                const GPlanet& p = getPlanet(i);
                if (config.RoundGravityWells()) {
                    if (distanceSquared(p.getPos(), pt) <= sqs)
                        return i;
                } else {
                    GPoint p2 = p.getPos().getSimpleNearestAlias(pt);
                    if (square(p2.x - pt.x) <= sqs && square(p2.y - pt.y) <= sqs)
                        return i;
                }
            }
        }
        return 0;
    } else {
        /* THost gravity wells: round, 3 ly, not wrapped, "cumulative" */
        int pid = 0;
        for (int i = ty_any_planets.findNextIndex(0); i != 0; i = ty_any_planets.findNextIndex(i)) {
            const GPlanet& p = getPlanet(i);
            if (distanceSquaredRaw(p.getPos(), pt) <= 9) {
                pt = p.getPos();  // (!)
                pid = i;
            }
        }
        return pid;
    }
}

/** Format name of a planet. */
static string_t
formatPlanetName(const GUniverse& univ, int pid, int flags)
{
    return format(((flags & locs_Orbit)
                   ? ((flags & locs_Verbose)
                      ? _("Orbit of %s (Planet #%d)")
                      : _("Orbit of %s (#%d)"))
                   : ((flags & locs_Verbose)
                      ? _("%s (Planet #%d)")
                      : _("%s (#%d)"))),
                  univ.getPlanet(pid).getName(GObject::PlainName), pid);
}

/** Get name of a location in human-readable form.
    \param pt     location
    \param flags  details of requested string
    - locs_Ships: also use ship names as location names, not just planets
    - locs_WW: return "near PLANETNAME" if applicable
    - locs_Orbit: say "Orbit of PLANET" if applicable
    - locs_Verbose: be more verbose */
string_t
GUniverse::getLocationName(GPoint pt, int flags) const
{
    /* is it a bird? */
    int pid = getPlanetAt(pt);
    if (pid)
        return formatPlanetName(*this, pid, flags);

    /* is it a plane? */
    if (flags & locs_Ships) {
        int sid = getAnyShipAt(pt);
        if (sid)
            return getShip(sid).getName(GObject::LongName);
    }
    if (flags & locs_WW) {
        pid = getGravityPlanetAt(pt);
        if (pid)
            return format(_("near %s"), formatPlanetName(*this, pid, flags & ~locs_Orbit));
    }
    if (flags & locs_NoSpace) {
        return string_t();
    }
    return format((flags & locs_Verbose) ? _("Deep Space (%d,%d)") : "(%d,%d)",
                  pt.x, pt.y);
}

/** Check whether a ship is being towed.
    \param sid Ship Id */
bool
GUniverse::isShipTowed(const int sid) const
{
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& sh = getShip(i);
        if (sid != i && sh.isPlayable(GObject::ReadOnly) && sh.getMission() == GMission::msn_Tow && sh.getTowId() == sid) {
            return true;
        }
    }
    return false;
}

/** Get ship at position. Any race does. This is mainly used for naming
    locations, and for things like to tell whether a "L" command would succeed.
    \param pt position to check
    \returns Id number of a ship at position \c pt, or zero if none. */
int
GUniverse::getAnyShipAt(GPoint pt) const
{
    pt = pt.getCanonicalLocation();
    for (int i = ty_any_ships.findNextIndex(0); i != 0; i = ty_any_ships.findNextIndex(i))
        if (getShip(i).getPos() == pt)
            return i;
    return 0;
}

/** Count own ships at position.
    \param pt position to check
    \returns number of currently-played ships which are at the
    specified position \c pt. */
int
GUniverse::getNumberOfPlayedShipsAt(GPoint pt) const
{
    int n = 0;
    pt = pt.getCanonicalLocation();
    for (int i = ty_played_ships.findNextIndex(0); i != 0; i = ty_played_ships.findNextIndex(i))
        if (getShip(i).getPos() == pt)
            ++n;
    return n;
}

/** Perform all updates. This will poll all updatable objects, and
    raise the appropriate signals. */
void
GUniverse::doScreenUpdates()
{
    bool changed = false;

    /* Tell everyone we're going to do updates */
    sig_pre_update.raise();

    /* Update individual objects */
    changed |= updateType(ty_history_ships);
    changed |= updateType(ty_any_planets);
    changed |= updateType(ty_ion_storms);
    changed |= updateType(ty_minefields);
    changed |= updateType(ty_ufos);

    /* Tell everyone we did updates */
    if (changed || universe_changed)
        sig_universe_changed.raise();
    universe_changed = false;
}

GPlanet&
GUniverse::getPlanet(int id)
{
    ASSERT(id > 0 && id <= NUM_PLANETS);
    return *planets[id-1];
}

const GPlanet&
GUniverse::getPlanet(int id) const
{
    ASSERT(id > 0 && id <= NUM_PLANETS);
    return *planets[id-1];
}

bool
GUniverse::isValidPlanetId(int id) const
{
    return id > 0 && id <= NUM_PLANETS;
}

GShip&
GUniverse::getShip(int id)
{
    ASSERT(id > 0 && id <= NUM_SHIPS);
    return *ships[id-1];
}

const GShip&
GUniverse::getShip(int id) const
{
    ASSERT(id > 0 && id <= NUM_SHIPS);
    return *ships[id-1];
}

bool
GUniverse::isValidShipId(int id) const
{
    return id > 0 && id <= NUM_SHIPS;
}

GIonStorm&
GUniverse::getIonStorm(int sid)
{
    ASSERT(sid > 0 && sid <= NUM_STORMS);
    return *storms[sid-1];
}

const GIonStorm&
GUniverse::getIonStorm(int sid) const
{
    ASSERT(sid > 0 && sid <= NUM_STORMS);
    return *storms[sid-1];
}

bool
GUniverse::isValidIonStormId(int id) const
{
    return id > 0 && id <= NUM_STORMS;
}

void
GUniverse::addMessageInformation(const GMessageInformation& msgi, int msgNr)
{
    switch (msgi.getObjectType()) {
     case MessageShip:
        if (isValidShipId(msgi.getObjectId())) {
            getShip(msgi.getObjectId()).addMessageInformation(msgi, msgNr);
        }
        break;
     case MessagePlanet:
     case MessageStarbase:
        if (isValidPlanetId(msgi.getObjectId()))
            getPlanet(msgi.getObjectId()).addMessageInformation(msgi, msgNr);
        break;
     case MessageMinefield:
        {
            // int id = msgi.getObjectId();
            TMinefieldReport report;
            report.x               = -1;
            report.y               = -1;
            report.owner           = -1;
            report.web             = false;
            report.units_known     = false;
            report.type_known      = false;
            report.radius_or_units = -1;
            report.turn            = getTurnNumber();
            report.why             = GMinefield::mfa_Scanned;
            for (GMessageInformation::iterator i = msgi.begin(); i != msgi.end(); ++i) {
                if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*i)) {
                    switch (iv->getIndex()) {
                     case mi_X:
                        report.x = iv->getValue();
                        break;
                     case mi_Y:
                        report.y = iv->getValue();
                        break;
                     case mi_Owner:
                        report.owner = iv->getValue();
                        break;
                     case mi_MineUnits:
                        report.radius_or_units = iv->getValue();
                        report.units_known = true;
                        break;
                     case mi_MineType:
                        report.web        = (iv->getValue() != 0);
                        report.type_known = true;
                        break;
                     case mi_MineScanReason:
                        report.why = iv->getValue();
                        break;
                     default:
                        break;
                    }
                }
            }

            /* Assimilate minefield when we have all necessary data */
            if (report.x > 0 && report.y > 0 && report.owner > 0 && report.radius_or_units >= 0) {
                ty_minefields.addMinefieldReport(msgi.getObjectId(), report);
            }
        }
        break;
     case MessageNoObject:
        break;

     case MessagePlayerScore:
        /* Score. Right now, we can only process current scores, because
           database access needs a timestamp. */
        if (isStatisticsFileLoaded() && msgi.getTurnNumber() == getTurnNumber())
            getStatistics().addMessageInformation(msgi, getTimestamp());
        break;

     case MessageConfig:
        for (GMessageInformation::iterator i = msgi.begin(); i != msgi.end(); ++i) {
            if (GMessageConfigValue* cv = dynamic_cast<GMessageConfigValue*>(*i)) {
                if (ConfigOption* opt = config.getConfigOption(cv->getIndex())) {
                    try {
                        opt->set(cv->getValue());
                    }
                    catch (...) {
                        /* Ignore any error. */
                    }
                }
            }
        }
        break;

     case MessageAlliance:
        /* Alliances do not go here, they go into the GAlliances object */
        break;

     case MessageIonStorm:
     case MessageExplosion:
        /* FIXME: implement me? */
        break;
    }
}

/** Initialize universe. This creates in all objects and hooks all signals. */
void
GUniverse::init()
{
    // Create objects
    for (int i = 0; i < NUM_PLANETS; ++i)
        planets.push_back(new GPlanet(i+1));
    for (int i = 0; i < NUM_SHIPS; ++i)
        ships.push_back(new GShip(i+1));
    for (int i = 0; i < NUM_STORMS; ++i)
        storms.push_back(new GIonStorm(i+1));

    // Observe signals
    // Not observed: ty_played/any/history_ships, ty_played/any_planets, ty_played_bases, ty_ion_storms
    ty_ufos.sig_set_changed.add(this, &GUniverse::markChanged);
    ty_fleets.sig_set_changed.add(this, &GUniverse::markChanged);
    ty_minefields.sig_set_changed.add(this, &GUniverse::markChanged);
}
