/**
  *  \file int/bytecode.cc
  *  \brief Interpreter: Bytecode Object
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/bytecode.h"
#include "cpluslib/assert.h"
#include "int/intglobal.h"
#include "int/expr.h"           // IntCompilationContext
#include "cpluslib/format.h"
#include "int/interror.h"


/** Find literal within data segment.
    \param dseg Data segment
    \param value Literal to find, must not be null
    \return index such that data[index] equals value. -1 if none found. */
static int32_t
findLiteral(const IntDataSegment& dseg, const IntValue* value)
{
    // Check at most 20 previous literals. This is to limit the amount of
    // time taken when compiling. As of 20100711, time taken and object file
    // sizes:                     core.q               selftest.q
    // - no literal merging    0.05s (28841 bytes)  0.08s (80985 bytes)
    // - max 20 literals       0.05s (25990 bytes)  0.11s (75894 bytes)
    // - full check            0.08s (24805 bytes)  0.90s (73186 bytes)
    // That is, we get roughly 2/3 of the savings at 1/30 of the cost.
    uint32_t first = 0;
    if (dseg.size() > 20)
        first = dseg.size() - 20;

    if (const IntIntValue* iv = dynamic_cast<const IntIntValue*>(value)) {
        for (uint32_t i = first; i < dseg.size(); ++i) {
            if (const IntIntValue* iv2 = dynamic_cast<const IntIntValue*>(dseg[i]))
                if (iv2->getValue() == iv->getValue() && iv2->isBool() == iv->isBool())
                    return i;
        }
        return -1;
    } else if (const IntFloatValue* fv = dynamic_cast<const IntFloatValue*>(value)) {
        for (uint32_t i = first; i < dseg.size(); ++i) {
            if (const IntFloatValue* fv2 = dynamic_cast<const IntFloatValue*>(dseg[i]))
                if (fv2->getValue() == fv->getValue())
                    return i;
        }
        return -1;
    } else if (const IntStringValue* sv = dynamic_cast<const IntStringValue*>(value)) {
        for (uint32_t i = first; i < dseg.size(); ++i) {
            if (const IntStringValue* sv2 = dynamic_cast<const IntStringValue*>(dseg[i]))
                if (sv2->getValue() == sv->getValue())
                    return i;
        }
        return -1;
    } else {
        return -1;
    }
}

static void addLineNumber(uint32_t line,
                          const std::vector<IntOpcode>& code,
                          std::vector<uint32_t>& line_numbers)
{
    uint32_t address = code.size();

    if (line_numbers.size() == 0
        || (line != line_numbers[line_numbers.size()-1] && address != line_numbers[line_numbers.size()-2]))
    {
        /* First pair, or new line at new address */
        line_numbers.push_back(address);
        line_numbers.push_back(line);
    } else if (address == line_numbers[line_numbers.size()-2]) {
        /* Same address as last pair, i.e. last line compiled to 0 instructions */
        line_numbers[line_numbers.size()-1] = line;
    } else {
        /* Same line as last pair, but different address, i.e. nested statement */
    }
}

static uint16_t verifyAddress(uint32_t arg)
{
    if (arg > 0xFFFF) {
        throw IntError("Code too complex");
    }
    return uint16_t(arg);
}

/** Construct blank BCO. */
IntBytecodeObject::IntBytecodeObject()
    : num_labels(0), min_args(0), max_args(0), is_procedure(true), is_varargs(false)
{ }

/** Destructor. */
IntBytecodeObject::~IntBytecodeObject()
{ }

/** Add named argument.
    \param name Name of argument
    \param optional true if this argument is optional

    addArgument(name, true) should not be followed by addArgument(name, false). */
void
IntBytecodeObject::addArgument(string_t name, bool optional)
{
    local_names.add(name);
    max_args = local_names.getNumNames();
    verifyAddress(max_args);
    if (!optional)
        min_args = local_names.getNumNames();
}

/** Get name of this BCO. Note that although this is a procedure name, it may not
    correspond with the current content of that procedure. */
string_t
IntBytecodeObject::getName() const
{
    return name;
}

/** Add a name for reference by later instructions. Existing names are recycled if
    possible. */
uint16_t
IntBytecodeObject::addName(string_t name)
{
    IntVariableNames::index_t idx = names.addMaybe(name);
    if (idx > 0xFFFF) {
        throw IntError("Code too complex (too many names)");
    }
    return uint16_t(idx);
}

/** Set name of this BCO. */
void
IntBytecodeObject::setName(string_t name)
{
    this->name = name;
}

/** Get file name of this BCO. */
string_t
IntBytecodeObject::getFileName() const
{
    return file_name;
}

/** Set file name of this BCO. */
void
IntBytecodeObject::setFileName(string_t file_name)
{
    this->file_name = file_name;
}

/** Notice current line number for debugging. */
void
IntBytecodeObject::addLineNumber(uint32_t line)
{
    ::addLineNumber(line, code, line_numbers);
}

/** Get line number for a particular PC. */
uint32_t
IntBytecodeObject::getLineNumber(pc_t pc) const
{
    /* Slow and simple */
    if (line_numbers.size() == 0 || pc < line_numbers[0])
        return 0;

    uint32_t i = 0;
    while (i+2 < line_numbers.size() && pc >= line_numbers[i+2])
        i += 2;
    return line_numbers[i+1];
}

/** Make a new label for future reference. This label can be used in as many jumps
    as needed (addJump), and must be placed exactly once using addLabel. */
IntBytecodeObject::label_t
IntBytecodeObject::makeLabel()
{
    if (num_labels >= 0xFFFF) {
        throw IntError("Code too complex (too many labels)");
    }
    return num_labels++;
}

/** Add an instruction. */
void
IntBytecodeObject::addInstruction(IntOpcode::Major major, uint8_t minor, uint16_t arg)
{
    IntOpcode o;
    o.major = major;
    o.minor = minor;
    o.arg   = arg;
    code.push_back(o);
}

void
IntBytecodeObject::addVerifyInstruction(IntOpcode::Major major, uint8_t minor, uint32_t arg)
{
    addInstruction(major, minor, verifyAddress(arg));
}

/** Add variable-referencing instruction.
    \param major Major opcode
    \param name  Name of referenced variable
    \param cc    Compilation context */
void
IntBytecodeObject::addVariableReferenceInstruction(IntOpcode::Major major, const string_t& name, const IntCompilationContext& cc)
{
    if (cc.compilation_flags & IntCompilationContext::LocalContext) {
        /* Is it a local variable? */
        IntVariableNames::index_t ix = local_names.lookup(name);
        if (ix != local_names.nil) {
            addVerifyInstruction(major, IntOpcode::sLocal, ix);
            return;
        }

        /* Is it a global variable? */
        if (cc.compilation_flags & IntCompilationContext::AlsoGlobalContext) {
            ix = global_property_names.lookup(name);
            if (ix != local_names.nil) {
                addVerifyInstruction(major, IntOpcode::sShared, ix);
                return;
            }
        }
    }

    /* Shortcut not possible */
    addInstruction(major, IntOpcode::sNamedVariable, addName(name));
}

/** Place a label. */
void
IntBytecodeObject::addLabel(label_t label)
{
    addInstruction(IntOpcode::maJump, IntOpcode::jSymbolic, label);
}

/** Add a jump. */
void
IntBytecodeObject::addJump(uint8_t flags, label_t label)
{
    addInstruction(IntOpcode::maJump, flags | IntOpcode::jSymbolic, label);
}

/** Add a "push literal" instruction. Checks whether the literal can be
    encoded as a single instruction; otherwise generates a new literal table slot. */
void
IntBytecodeObject::addPushLiteral(IntValue* literal)
{
    /* Is it empty? */
    if (literal == 0) {
        addInstruction(IntOpcode::maPush, IntOpcode::sBoolean, uint16_t(-1U));
        return;
    }

    /* Is it a small integer literal? */
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(literal)) {
        if (iv->getValue() >= -int32_t(0x7FFF) && iv->getValue() <= int32_t(0x7FFF)) {
            if (iv->isBool()) {
                addInstruction(IntOpcode::maPush, IntOpcode::sBoolean, iv->getValue());
            } else {
                addInstruction(IntOpcode::maPush, IntOpcode::sInteger, iv->getValue());
            }
            return;
        }
    }

    /* None of the above, so use general way */
    int32_t existing = findLiteral(data, literal);
    if (existing >= 0) {
        // Recycle existing literal
        addInstruction(IntOpcode::maPush, IntOpcode::sLiteral, existing);
    } else {
        if (data.size() >= 0xFFFF) {
            throw IntError("Code too complex (too many literals)");
        }
        data.push(literal);
        addInstruction(IntOpcode::maPush, IntOpcode::sLiteral, data.size()-1);
    }
}

/** Check whether this BCO contains any instruction that potentially calls user code.
    This can be
    - callind, pushind, procind (for convenience, any maIndirect)
    - sevalx
    - sevals
    - srunhook */
bool
IntBytecodeObject::hasUserCall() const
{
    for (pc_t i = 0; i != code.size(); ++i) {
        if (code[i].major == IntOpcode::maIndirect
            || (code[i].major == IntOpcode::maSpecial
                && (code[i].minor == IntOpcode::miSpecialEvalStatement
                    || code[i].minor == IntOpcode::miSpecialEvalExpr
                    || code[i].minor == IntOpcode::miSpecialRunHook)))
        {
            return true;
        }
    }
    return false;
}

/** Turn symbolic references into absolute references. Also eliminates
    null operations (absolute labels). */
void
IntBytecodeObject::relocate()
{
    std::vector<pc_t> addresses(num_labels, pc_t(code.size()));

    // Find existing labels
    pc_t outAdr = 0;
    pc_t lastAdr = 0;
    for (pc_t i = 0; i != code.size(); ++i) {
        if (code[i].isLabel()) {
            /* It's a label. Those do not produce output. */
            if (code[i].minor & IntOpcode::jSymbolic) {
                /* Symbolic label. Note its address. */
                ASSERT(code[i].arg < addresses.size());
                addresses[code[i].arg] = outAdr;
                lastAdr = outAdr;
            } else {
                /* Absolute label aka NOP */
            }
        } else {
            ++outAdr;
        }
    }
    
    if (lastAdr > 0xFFFF) {
        // Unrelocatable
        return;
    }

    // Turn symbolic jumps into absolute.
    std::vector<IntOpcode> newCode;
    std::vector<uint32_t> newDebug;
    uint32_t dbgIndex = 0;
    for (pc_t i = 0; i != code.size(); ++i) {
        /* Update debug information */
        if (dbgIndex < line_numbers.size() && line_numbers[dbgIndex] == i) {
            ::addLineNumber(line_numbers[dbgIndex+1], newCode, newDebug);
            dbgIndex += 2;
        }
        /* Update code */
        if (code[i].major == IntOpcode::maJump) {
            if (code[i].isLabel()) {
                /* Label. Drop it. */
            } else if ((code[i].minor & IntOpcode::jSymbolic) != 0) {
                /* Make it absolute */
                ASSERT(code[i].arg < addresses.size());
                IntOpcode c;
                c.major = code[i].major;
                c.minor = code[i].minor & ~IntOpcode::jSymbolic;
                c.arg   = addresses[code[i].arg];
                newCode.push_back(c);
            } else {
                /* Already absolute? Should not happen. */
                newCode.push_back(code[i]);
            }
        } else {
            newCode.push_back(code[i]);
        }
    }

    code.swap(newCode);
    line_numbers.swap(newDebug);
}

/** Compact code. This drops all NOPs. This is a subset of relocate()
    used for optimisation. */
void
IntBytecodeObject::compact()
{
    std::vector<IntOpcode> oldCode;
    std::vector<uint32_t> oldDebug;
    code.swap(oldCode);
    line_numbers.swap(oldDebug);
    uint32_t dbgIndex = 0;
    for (pc_t i = 0; i != oldCode.size(); ++i) {
        /* Update debug information */
        if (dbgIndex < oldDebug.size() && oldDebug[dbgIndex] == i) {
            addLineNumber(oldDebug[dbgIndex+1]);
            dbgIndex += 2;
        }
        /* Update code */
        if (oldCode[i].major == IntOpcode::maJump && oldCode[i].minor == IntOpcode::jLabel) {
            /* Absolute label aka NOP. Drop it. */
        } else {
            /* Instruction. Keep it. */
            code.push_back(oldCode[i]);
        }
    }
}

/** Copy local variables from another BCO. */
void
IntBytecodeObject::copyLocals(const IntBytecodeObject& other)
{
    for (IntVariableNames::index_t i = 0, e = other.local_names.getNumNames(); i != e; ++i) {
        local_names.add(other.local_names.lookupAddress(i));
    }
}

/** Append another byte-code object.
    Note that this trusts the BCO to contain no invalid opcodes. */
void
IntBytecodeObject::append(IntBytecodeObject& other)
{
    // Remember base address of insertion
    pc_t absBase = code.size();
    label_t symBase = num_labels;

    if (other.num_labels > 0xFFFF - num_labels) {
        throw IntError("Code too complex (too many labels)");
    }

    num_labels += other.num_labels;
    code.reserve(code.size() + other.code.size());

    // Copy the code
    for (pc_t i = 0; i != other.code.size(); ++i) {
        const IntOpcode& o = other.code[i];
        switch (const IntOpcode::Major maj = IntOpcode::Major(o.major)) {
         case IntOpcode::maPush:
         case IntOpcode::maPop:
         case IntOpcode::maStore:
         case IntOpcode::maFusedUnary:
         case IntOpcode::maFusedBinary:
         case IntOpcode::maFusedComparison2:
         case IntOpcode::maInplaceUnary:
            // Handle scope
            switch (IntOpcode::Scope(o.minor)) {
             case IntOpcode::sNamedVariable:
             case IntOpcode::sNamedShared:
                // Adjust name reference
                addInstruction(maj, o.minor, addName(other.getName(o.arg)));
                break;
             case IntOpcode::sLocal:
                // Adjust local by name
                // FIXME: this works as long as all locals are unique. It will fail
                // when we have non-unique locals as there is no requirement that the
                // local addresses used in the code correspond to local_names. We
                // should make up some rules about how the two BCOs have to be related.
                addVerifyInstruction(maj, o.minor, local_names.addMaybe(other.local_names.lookupAddress(o.arg)));
                break;
             case IntOpcode::sLiteral:
                // Adjust literal
                addPushLiteral(other.data[o.arg]);
                break;
             case IntOpcode::sInteger:
             case IntOpcode::sBoolean:
             case IntOpcode::sStatic:
             case IntOpcode::sShared:
                // Copy verbatim
                code.push_back(o);
                break;
            }
            break;
         case IntOpcode::maBinary:
         case IntOpcode::maUnary:
         case IntOpcode::maTernary:
         case IntOpcode::maStack:
         case IntOpcode::maIndirect:
         case IntOpcode::maFusedComparison:
            // Copy verbatim
            code.push_back(o);
            break;
         case IntOpcode::maJump:
            // Adjust argument
            if (o.minor & IntOpcode::jSymbolic) {
                addInstruction(maj, o.minor, o.arg + symBase);
            } else {
                addVerifyInstruction(maj, o.minor, o.arg + absBase);
            }
            break;
         case IntOpcode::maMemref:
         case IntOpcode::maDim:
            // Adjust name reference
            addInstruction(maj, o.minor, addName(other.getName(o.arg)));
            break;
         case IntOpcode::maSpecial:
            switch (IntOpcode::Special(o.minor)) {
             case IntOpcode::miSpecialUncatch:
             case IntOpcode::miSpecialReturn:
             case IntOpcode::miSpecialWith:
             case IntOpcode::miSpecialEndWith:
             case IntOpcode::miSpecialFirstIndex:
             case IntOpcode::miSpecialNextIndex:
             case IntOpcode::miSpecialEndIndex:
             case IntOpcode::miSpecialEvalStatement:
             case IntOpcode::miSpecialEvalExpr:
             case IntOpcode::miSpecialLoad:
             case IntOpcode::miSpecialPrint:
             case IntOpcode::miSpecialAddHook:
             case IntOpcode::miSpecialRunHook:
             case IntOpcode::miSpecialThrow:
             case IntOpcode::miSpecialTerminate:
             case IntOpcode::miSpecialSuspend:
             case IntOpcode::miSpecialNewArray:
             case IntOpcode::miSpecialMakeList:
             case IntOpcode::miSpecialNewHash:
             case IntOpcode::miSpecialInstance:
             case IntOpcode::miSpecialResizeArray:
             case IntOpcode::miSpecialBind:
             case IntOpcode::miSpecialFirst:
             case IntOpcode::miSpecialNext:
                // Copy verbatim
                code.push_back(o);
                break;
             case IntOpcode::miSpecialDefSub:
             case IntOpcode::miSpecialDefShipProperty:
             case IntOpcode::miSpecialDefPlanetProperty:
                // Adjust name reference
                addInstruction(maj, o.minor, addName(other.getName(o.arg)));
                break;
            }
            break;
        }
    }

    // FIXME: copy debug information. Right now, this is not needed because there
    // is no debug information in BCOs we merge.
}

/** Find jump target. If the jump is symbolic, looks up the target label.
    \param minor Minor opcode from the maJump instruction
    \param arg Parameter */
IntBytecodeObject::pc_t
IntBytecodeObject::getJumpTarget(uint8_t minor, uint16_t arg) const
{
    if (minor & IntOpcode::jSymbolic) {
        // Symbolic jump: find label
        for (pc_t i = 0; i != code.size(); ++i)
            if (code[i].major == IntOpcode::maJump && code[i].minor == (IntOpcode::jSymbolic + IntOpcode::jLabel))
                if (code[i].arg == arg)
                    return i;
        return code.size();
    } else {
        return arg;
    }
}

/** Get the human-readable value of the instruction at address \c index. */
string_t
IntBytecodeObject::getDisassembly(pc_t index) const
{
    string_t tpl;
    code[index].getDisassemblyTemplate(tpl);

    string_t result;
    for (string_t::size_type i = 0; i < tpl.size(); ++i) {
        if (tpl[i] == '\t') {
            result.append(12 - (result.size()%12), ' ');
        } else if (tpl[i] == '%' && i+1 < tpl.size()) {
            char mode = tpl[++i];
            uint16_t arg = code[index].arg;

            /* Append arg in raw form */
            if (mode == 'd') {
                result += format("%d", int16_t(arg));
            } else {
                result += format("%d", arg);
            }

            /* If we have a hint, append that as well */
            switch (mode) {
             case 'n':
                result += " <";
                result += getName(arg);
                result += ">";
                break;
             case 'l': /* literal */
                result += " <";
                result += IntValue::toStringOf(data[arg], true);
                result += ">";
                break;
             case 'd': /* signed */
             case 'u': /* unsigned */
             case 'T': /* static given by address */
                break;
             case 'G': /* shared given by address */
                if (arg < global_property_names.getNumNames()) {
                    result += " <";
                    result += global_property_names.lookupAddress(arg);
                    result += ">";
                }
                break;
             case 'L': /* local given by address */
                if (arg < local_names.getNumNames()) {
                    result += " <";
                    result += local_names.lookupAddress(arg);
                    result += ">";
                }
                break;
            }
        } else {
            result += tpl[i];
        }
    }

    return result;
}
