/**
  *  \file reshack/reshack.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <vector>
#include "arch/platform.h"
#include "cpluslib/ptrvector.h"
#include "gfx/bmfont.h"
#include "gfx/init.h"
#include "gfx/palette.h"
#include "reshack/edit.h"
#include "reshack/fontedit.h"
#include "reshack/picedit.h"
#include "reshack/tools.h"
#include "reshack/load.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "util/cmdline.h"
#include "util/console.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "reshack/charname.h"
#include "reshack/colors.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/focusiterator.h"
#include "util/misc.h"
#include "ui/dialogs/filedlg.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"
#include "ui/widgets/radiobutton.h"
#ifdef USE_W32IMPORT
# include "reshack/w32import.h"
#endif

namespace {
    /** List of editors. */
    class RHEditorList : public UIListbox {
        ptr_vector<RHEditor>& editors;
     public:
        RHEditorList(ptr_vector<RHEditor>& editors);
        void drawPart(GfxCanvas& can, int from, int to);
        void update();
    };

    /** Resource hacker main dialog. Displays a list of editors
        (RHEditorList) and lets the user add and manipulate them. */
    class RHMainDialog : public UIWindow {
        ObjectHolder h;
        ptr_vector<RHEditor> editors;
        RHEditorList list;

        void init();
        void onEdit();
        void onLoad();
        void onSave();
        void onExit();

        void addNewEditor(RHEditor* ed);
     public:
        RHMainDialog();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);

        void doLoad(string_t filename);
    };

    /** New Picture dialog. */
    class RHNewPictureDialog : public UIWindow {
        ObjectHolder h;
        UIInputLine* input_width;
        UIInputLine* input_height;
        Observable<int> kind;
        GfxPoint size;

        void init();
        void onOK();
     public:
        RHNewPictureDialog();
        GfxPoint getSize() const
            { return size; }
        int getKind() const
            { return kind.get(); }
    };
}

/****************************** RHEditorList *****************************/

/** Constructor.
    \param editors List to display */
RHEditorList::RHEditorList(ptr_vector<RHEditor>& editors)
    : UIListbox(0, 15, font_ems[FONT_NORMAL]*30, editors.size()),
      editors(editors)
{ }

void
RHEditorList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        uint32_t ele = getTop() + i;
        string_t line;
        if (ele < editors.size()) {
            line += editors[ele]->getName();
            if (editors[ele]->hasUnsavedChanges())
                line += " *";
        }
        outTextF(ctx, area.x, area.y, area.w, line);
    }
}

/** Update list. Call after manipulating the list. */
void
RHEditorList::update()
{
    setSize(editors.size());
    drawWidget();
}


/****************************** RHMainDialog *****************************/

RHMainDialog::RHMainDialog()
    : UIWindow(_("Resource Editor"), BLUE_WINDOW, COLOR_GRAY),
      list(editors)
{
    init();
}

/** Create dialog. */
void
RHMainDialog::init()
{
    // HBox
    //   RHEditorList
    //   VBox
    //     UIButton "Edit"
    //     UIButton "Load"
    //     UIButton "Save"
    //     UIButton "New"
    //     UISpacer
    //     UIButton "Exit"
    
    setLayout(UIHBoxLayout::instance5);
    add(list);

    UIButton& btn_edit = h.add(new UIButton(1000, _("Edit"), 'e'));
    UIButton& btn_load = h.add(new UIButton(1001, _("Load"), 'l'));
    UIButton& btn_save = h.add(new UIButton(1002, _("Save"), 's'));
    UIButton& btn_new  = h.add(new UIButton(1003, _("New"),  'n'));
    UIButton& btn_exit = h.add(new UIButton(1004, _("Exit"), SDLK_ESCAPE));

    btn_edit.sig_fired.add(this, &RHMainDialog::onEdit);
    btn_load.sig_fired.add(this, &RHMainDialog::onLoad);
    btn_save.sig_fired.add(this, &RHMainDialog::onSave);
    btn_new.addCommand();
    btn_exit.sig_fired.add(this, &RHMainDialog::onExit);

    UIGroup& v = h.add(new UIGroup(UIVBoxLayout::instance5));
    v.add(btn_edit);
    v.add(btn_load);
    v.add(btn_save);
    v.add(btn_new);
    v.add(h.add(new UISpacer()));
    v.add(btn_exit);
    add(v);

    pack();
}

/** "Edit" function. Calls the current editor's edit function. */
void
RHMainDialog::onEdit()
{
    if (list.getCurrent() >= 0 && list.getCurrent() < int(editors.size())) {
        editors[list.getCurrent()]->edit();
        list.updateCurrent();
    }
}

/** "Load" function. Asks for a file name and creates a new editor displaying that file. */
void
RHMainDialog::onLoad()
{
    // Ask user for file name
    UIFileChooserDialog flw(_("Load"), Directory::openDir(Directory::getWorkingDirectory()), "*");
    if (!ui_root->runModal(flw)) {
        return;
    }

    // Load file
    doLoad(flw.getResultFile());
}

/** "Save" function. Calls the current editor's save function. */
void
RHMainDialog::onSave()
{
    if (list.getCurrent() >= 0 && list.getCurrent() < int(editors.size())) {
        editors[list.getCurrent()]->save();
        list.updateCurrent();
    }
}

/** "Exit" function. */
void
RHMainDialog::onExit()
{
    // Any unsaved changes?
    bool unsavedChanges = false;
    for (ptr_vector<RHEditor>::iterator i = editors.begin(); i != editors.end(); ++i)
        if ((*i)->hasUnsavedChanges())
            unsavedChanges = true;

    // Prompt
    if (!unsavedChanges || UIMessageBox(_("Do you want to exit? There are unsaved changes."), _("Resource Editor")).doYesNoDialog())
        stop(1);
}

/** Add a new editor. Takes ownership of the editor.
    \param ed the editor */
void
RHMainDialog::addNewEditor(RHEditor* ed)
{
    editors.push_back(ed);
    list.update();
    list.scrollTo(editors.size()-1, 1);
}

bool
RHMainDialog::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == 1003) {
        StringList sl;
        sl.add(1, _("New Picture"));
        if (getClipboardImage())
            sl.add(2, _("New Picture from Clipboard"));
        sl.add(3, _("New Font"));
#ifdef USE_W32IMPORT
        sl.add(4, _("Import system font"));
#endif

        UIStandardListbox lb(0, sl.getCount(), 20 + getTextWidth(*fonts[FONT_NORMAL], sl), sl);
        if (!lb.doStandardMenu(sender))
            return true;

        switch (lb.getKey()) {
         case 1:
            {
                RHNewPictureDialog dlg;
                if (ui_root->runModal(dlg)) {
                    Ptr<GfxPixmap> pix;
                    if (dlg.getKind() == TrueColor) {
                        pix = new GfxPixmap(dlg.getSize().x, dlg.getSize().y, 32, 0xFF0000, 0x00FF00, 0x0000FF, 0);
                    } else {
                        pix = new GfxPixmap(dlg.getSize().x, dlg.getSize().y, 8, 0, 0, 0, 0);
                        SDL_SetPalette(pix->getSystemSurface(), SDL_LOGPAL|SDL_PHYSPAL, standard_palette, 0, NUM_STANDARD_COLORS);
                    }
                    addNewEditor(new RHPictureEditor(pix, ColorMode(dlg.getKind())));
                }
            }
            break;

         case 2:
            addNewEditor(new RHPictureEditor(getClipboardImage()->clone(), ""));
            break;

         case 3:
            {
                UIDecimalSelector ds(0, 1, 100, 10);
                if (ds.doStandardDialog(_("New Font"), _("Font Height:"), string_t())) {
                    Ptr<GfxBitmapFont> bm = new GfxBitmapFont();
                    bm->addNewGlyph(0, new GfxBitmapGlyph(1, ds.getValue()));
                    bm->addNewGlyph(0, 0);
                    addNewEditor(new RHFontEditor(bm, ""));
                }
            }
            break;

#ifdef USE_W32IMPORT
         case 4:
            if (Ptr<GfxBitmapFont> bm = importWin32Font()) {
                addNewEditor(new RHFontEditor(bm, ""));
            }
            break;
#endif
        }
        return true;
    } else {
        return UIWindow::handleCommand(cmd, sender);
    }
}

void
RHMainDialog::doLoad(string_t filename)
{
    // Open file
    Ptr<Stream> s = Stream::openFileNT(filename, Stream::C_READ);
    if (!s) {
        messageBox(_("Unable to open file"), _("Load"));
        return;
    }

    // Try to load picture
    Ptr<GfxPixmap> pix;
    try {
        pix = loadImage(*s);
    }
    catch (...) { }

    // When picture successfully loaded, finish.
    if (pix) {
        addNewEditor(new RHPictureEditor(pix, filename));
        return;
    }

    // Try to load font
    Ptr<GfxBitmapFont> font = new GfxBitmapFont();
    try {
        s->seek(0);
        font->load(*s, 0);
        addNewEditor(new RHFontEditor(font, filename));
        return;
    }
    catch (...) { }

    // Try to load as BDF
    font = new GfxBitmapFont();
    s->seek(0);
    if (loadBDF(*font, *s)) {
        addNewEditor(new RHFontEditor(font, filename));
        return;
    }

    // Nothing worked
    messageBox(_("Unable to load file: ") + filename, _("Load"));
}

/**************************** RHNewPictureDialog ***************************/

RHNewPictureDialog::RHNewPictureDialog()
    : UIWindow(_("New Picture"), BLUE_WINDOW, COLOR_GRAY),
      input_width(0),
      input_height(0),
      kind(StandardPaletteColor),
      size(0, 0)
{
    init();
}

void
RHNewPictureDialog::init()
{
    // VBox
    //   Grid
    //     "Width" UIInputLine
    //     "Height" UIInputLine
    //   "Format"
    //   UIRadioButton*
    //   HBox
    //     UISpacer
    //     "OK"
    //     "Cancel"

    UIGroup& g1 = h.add(new UIGroup(h.add(new UIGridLayout(2))));
    UIGroup& g4 = h.add(new UIGroup(UIHBoxLayout::instance5));

    input_width  = &h.add(new UIInputLine(10, ss_Alt + 'w', 0, 1000));
    input_height = &h.add(new UIInputLine(10, ss_Alt + 'h', 0, 1001));
    g1.add(h.add(new UIStaticText(_("Width:"), UIStaticText::StaticTitle)));
    g1.add(*input_width);
    g1.add(h.add(new UIStaticText(_("Height:"), UIStaticText::StaticTitle)));
    g1.add(*input_height);
    add(g1);

    int max;
    if (ui_screen_bits <= 8) {
        // No choice in color format
        max = 1001;
    } else {
        // Color format choices
        add(h.add(new UIStaticText(_("Color format:"), UIStaticText::StaticTitle)));
        add(h.add(new UIRadioButton(1002, _("Standard palette"), ss_Alt + 's', kind, StandardPaletteColor)));
        add(h.add(new UIRadioButton(1003, _("Foreign palette"), ss_Alt + 'f', kind, ForeignPaletteColor)));
        add(h.add(new UIRadioButton(1004, _("Truecolor"), ss_Alt + 't', kind, TrueColor)));
        max = 1004;
    }
    add(h.add(new UIFocusIterator(1000, max, UIFocusIterator::fi_Vert + UIFocusIterator::fi_Tab)));

    g4.add(h.add(new UISpacer()));
    g4.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &RHNewPictureDialog::onOK));
    g4.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    add(g4);

    setLayout(UIVBoxLayout::instance5);
    pack();
    ui_root->centerWidget(*this);
}

void
RHNewPictureDialog::onOK()
{
    int width, height;
    if (!stringValue(input_width->getText(), width) || width <= 0 || width > 2500) {
        input_width->setFocusedWidget();
        return;
    }
    if (!stringValue(input_height->getText(), height) || height <= 0 || height > 2000) {
        input_height->setFocusedWidget();
        return;
    }
    size.x = width;
    size.y = height;
    stop(1);
}

/************************** Command Line Parser **************************/

class RHArgs : public StandardCommandLineParser {
 public:
    RHArgs()
        { }
    void handleNonoption(const string_t& x);
    void handleError(const string_t& x);

    // public for simplicity
    std::vector<string_t> names;
};

void
RHArgs::handleNonoption(const string_t& x)
{
    names.push_back(x);
}

void
RHArgs::handleError(const string_t& x)
{
    console.write(LOG_ERROR, x);
    Platform::showOutput();
    std::exit(1);
}


static const ModuleInitializer*const reshack_modules[] = {
    mod_base_gui,
    0,
};

int main(int /*argc*/, char** argv)
{
    performInitialisation(OnStartup, reshack_modules);

    /* Parse args */
    RHArgs parser;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    parser.parse(Platform::convertArgumentsToUtf8(argv));

    /* Initialize */
    try {
        initUIGlobals(0, "PCC2 Resource Hack");
        loadCharacterNames();
    }
    catch (std::exception& e) {
        console.write(LOG_ERROR, string_t(_("Error during initialisation: ")) + e.what());
        Platform::showOutput();
        return 1;
    }

    /* GUI */
    RHMainDialog main;
    ui_root->centerWidget(main);
    ui_root->add(main);
    for (std::vector<string_t>::iterator i = parser.names.begin(); i != parser.names.end(); ++i)
        main.doLoad(*i);
    ui_root->run();
    ui_root->close(main);
    return 0;
}
