/**
  *  \file u/t_util_conf.cc
  *  \brief Test for util/conf.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/conf.h"
#include "u/t_util.h"
#include "cxxtest/TestSuite.h"
#include "util/except.h"

void
UtilConfTestSuite::testValueIntParser()
{
    ValueIntParser vip;

    // Some random values
    TS_ASSERT_EQUALS(vip.parse("0"), 0);
    TS_ASSERT_EQUALS(vip.parse("1"), 1);
    TS_ASSERT_EQUALS(vip.parse("65535"), 65535);
    TS_ASSERT_EQUALS(vip.parse("65536"), 65536);
    TS_ASSERT_EQUALS(vip.parse("2147483647"), 2147483647);
    TS_ASSERT_EQUALS(vip.parse("-1"), -1);
    TS_ASSERT_EQUALS(vip.parse("-2147483648"), int32(-2147483648U));

    // Spacing etc.
    TS_ASSERT_EQUALS(vip.parse(" 42"), 42);
    TS_ASSERT_EQUALS(vip.parse(" 42      "), 42);
    TS_ASSERT_EQUALS(vip.parse("42        "), 42);

    // Wrong stuff
    TS_ASSERT_THROWS(vip.parse("x"), std::range_error);
    TS_ASSERT_THROWS(vip.parse("x42"), std::range_error);
    // no longer wrong, as we want to parse things like '100%':
    // TS_ASSERT_THROWS(vip.parse("42x"), std::range_error);

    // Reverse conversion
    TS_ASSERT_EQUALS(vip.toString(0), "0");
    TS_ASSERT_EQUALS(vip.toString(1), "1");
    TS_ASSERT_EQUALS(vip.toString(65535), "65535");
    TS_ASSERT_EQUALS(vip.toString(65536), "65536");
    TS_ASSERT_EQUALS(vip.toString(2147483647), "2147483647");
    TS_ASSERT_EQUALS(vip.toString(-1), "-1");
    TS_ASSERT_EQUALS(vip.toString(-2147483648U), "-2147483648");
}

void
UtilConfTestSuite::testValueBoolParser()
{
    ValueBoolParser vbp;

    // Test some values
    TS_ASSERT_EQUALS(vbp.parse("no"), 0);
    TS_ASSERT_EQUALS(vbp.parse("yes"), 1);
    TS_ASSERT_EQUALS(vbp.parse("allies"), 2);
    TS_ASSERT_EQUALS(vbp.parse("external"), 3);
    TS_ASSERT_EQUALS(vbp.parse("true"), 1);
    TS_ASSERT_EQUALS(vbp.parse("false"), 0);

    // Short forms
    TS_ASSERT_EQUALS(vbp.parse("n"), 0);
    TS_ASSERT_EQUALS(vbp.parse("y"), 1);
    TS_ASSERT_EQUALS(vbp.parse("a"), 2);
    TS_ASSERT_EQUALS(vbp.parse("e"), 3);
    TS_ASSERT_EQUALS(vbp.parse("t"), 1);
    TS_ASSERT_EQUALS(vbp.parse("f"), 0);

    // Case
    TS_ASSERT_EQUALS(vbp.parse("NO"), 0);
    TS_ASSERT_EQUALS(vbp.parse("YES"), 1);
    TS_ASSERT_EQUALS(vbp.parse("ALL"), 2);
    TS_ASSERT_EQUALS(vbp.parse("EXT"), 3);
    TS_ASSERT_EQUALS(vbp.parse("TRU"), 1);
    TS_ASSERT_EQUALS(vbp.parse("FAL"), 0);

    // Reverse conversion
    TS_ASSERT_EQUALS(vbp.toString(0), "No");
    TS_ASSERT_EQUALS(vbp.toString(1), "Yes");
    TS_ASSERT_EQUALS(vbp.toString(2), "Allies");
    TS_ASSERT_EQUALS(vbp.toString(3), "External");
}

void
UtilConfTestSuite::testValueBitsetParser()
{
    ValueBitsetParser vbp("one,two,three,four,five");

    // Single values
    TS_ASSERT_EQUALS(vbp.parse(""), 0);
    TS_ASSERT_EQUALS(vbp.parse("one"), 1);
    TS_ASSERT_EQUALS(vbp.parse("two"), 2);
    TS_ASSERT_EQUALS(vbp.parse("three"), 4);
    TS_ASSERT_EQUALS(vbp.parse("four"), 8);
    TS_ASSERT_EQUALS(vbp.parse("five"), 16);

    // Multiple values
    TS_ASSERT_EQUALS(vbp.parse("one,two"), 3);
    TS_ASSERT_EQUALS(vbp.parse("two,three,four"), 14);
    TS_ASSERT_EQUALS(vbp.parse("five,three"), 20);
    TS_ASSERT_EQUALS(vbp.parse("one,one,one,one"), 1);
    TS_ASSERT_EQUALS(vbp.parse("five,,,,,,,,"), 16);

    // Numerical values
    TS_ASSERT_EQUALS(vbp.parse("one,120"), 121);
    TS_ASSERT_EQUALS(vbp.parse("one,121"), 121);
    TS_ASSERT_EQUALS(vbp.parse("121,one"), 121);

    // Reverse conversion
    TS_ASSERT_EQUALS(vbp.toString(0), "");
    TS_ASSERT_EQUALS(vbp.toString(1), "one");
    TS_ASSERT_EQUALS(vbp.toString(2), "two");
    TS_ASSERT_EQUALS(vbp.toString(3), "one,two");
    TS_ASSERT_EQUALS(vbp.toString(4), "three");
    TS_ASSERT_EQUALS(vbp.toString(32), "");
}
